# standard library
from typing import Union, Optional, TYPE_CHECKING

# third party imports
import arviz as az
import numpy as np
import matplotlib.pyplot as plt
from loguru import logger

# local imports
from probeye.subroutines import len_or_one
from probeye.subroutines import add_index_to_tex_prm_name

# imports only needed for type hints
if TYPE_CHECKING:  # pragma: no cover
    from probeye.definition.inference_problem import InferenceProblem


def create_pair_plot(
    inference_data: az.data.inference_data.InferenceData,
    problem: "InferenceProblem",
    plot_with: str = "arviz",
    plot_priors: bool = True,
    focus_on_posterior: bool = False,
    kind: str = "kde",
    figsize: Optional[tuple] = (9, 9),
    textsize: Union[int, float] = 10,
    title_size: Union[int, float] = 14,
    title: Optional[str] = None,
    true_values: Optional[dict] = None,
    show_legends: bool = True,
    show: bool = True,
    **kwargs,
) -> np.ndarray:
    """
    Creates a pair-plot for the given inference data.

    Parameters
    ----------
    inference_data
        Contains the results of the sampling procedure.
    problem
        The inference problem the inference data refers to.
    plot_with
        Defines the python package the plot will be generated with. Options are:
        {'arviz', 'seaborn', 'matplotlib'}.
    plot_priors
        If True, the prior-distributions are included in the marginal subplots.
        Otherwise the priors are not shown.
    focus_on_posterior
        If True, the marginal plots will focus on the posteriors, i.e., the range of the
        horizontal axis will adapt to the posterior. This might result in just seeing a
        fraction of the prior distribution (if they are included). If False, the
        marginal plots will focus on the priors, which will have a broader x-range. If
        plot_priors=False, this argument has no effect on the generated plot.
    kind
        Type of plot to display ('scatter', 'kde' and/or 'hexbin').
    figsize
        Defines the size of the generated plot in the default unit. If None is chosen,
        the figsize will be derived automatically.
    textsize
        Defines the font size in the default unit.
    title_size
        Defines the font size of the figures title if 'title' is given.
    title
        The title of the figure.
    true_values
        Used for plotting 'true' parameter values. Keys are the parameter names and
        values are the values that are supposed to be shown in the marginal plots.
    show_legends
        If True, legends are shown in the marginal plots. Otherwise no legends are
        included in the plot.
    show
        When True, the show-method is called after creating the plot. Otherwise, the
        show-method is not called. The latter is useful, when the plot should be further
        processed.
    kwargs
        Additional keyword arguments passed to arviz' pairplot function.

    Returns
    -------
    axs
        The array of subplots of the created plot.
    """

    # a pairplot can only be generate when there are at least two parameter or parameter
    # components (the latter refers to vector-valued parameters)
    if problem.n_latent_prms_dim == 1:
        logger.warning(
            "The combined dimension of all latent parameters is one. Hence, no "
            "pairplot can be generated in this setup."
        )
        return np.array([])

    if plot_with == "arviz":

        # set default value for kde_kwargs if not given in kwargs; note that this
        # default value is mutable, so it should not be given as a default argument in
        # create_pair_plot
        if "kde_kwargs" not in kwargs:
            kwargs["kde_kwargs"] = {
                "contourf_kwargs": {"alpha": 0},
                "contour_kwargs": {"colors": None},
            }

        # process true_values if specified
        if true_values is not None:
            reference_values = dict()
            for prm_name, value in true_values.items():
                dim = problem.parameters[prm_name].dim
                tex = problem.parameters[prm_name].tex
                if dim > 1:
                    if tex in inference_data.posterior.keys():
                        # in this case, the inference_data object contains samples of a
                        # multidimensional parameter; the corresponding reference values
                        # must be given in a dict <tex>\n0, <tex>\n1, <tex>\n2, etc.
                        for i in range(dim):
                            key = f"{tex}\n{i}"
                            reference_values[key] = value[i]
                    else:
                        # in this case, the samples of a multidimensional parameter have
                        # been already decomposed, so all the channels in the inference
                        # data are 1D
                        for i in range(dim):
                            key = add_index_to_tex_prm_name(tex, i + 1)
                            reference_values[key] = value[i]
                else:
                    key = tex
                    reference_values[key] = value
            kwargs["reference_values"] = reference_values
            if "reference_values_kwargs" not in kwargs:
                kwargs["reference_values_kwargs"] = {"marker": "o", "color": "red"}

        # call the main plotting routine from arviz
        axs = az.plot_pair(
            inference_data,
            marginals=True,
            kind=kind,
            figsize=figsize,
            textsize=textsize,
            show=show,
            **kwargs,
        )

        # adds a reference value in each marginal plot; for some reason this is not done
        # by arviz.pair_plot when passing 'reference_values'
        if "reference_values" in kwargs:
            reference_values_kwargs = None
            if "reference_values_kwargs" in kwargs:
                reference_values_kwargs = kwargs["reference_values_kwargs"]
            ref_value_list = [*kwargs["reference_values"].values()]
            if problem.n_latent_prms_dim > 2:
                # in this case, the relevant axis is always the horizontal one
                for i, prm_value in enumerate(ref_value_list):
                    axs[i, i].scatter(
                        prm_value,
                        0,
                        label="true value",
                        zorder=10,
                        **reference_values_kwargs,
                        edgecolor="black",
                    )
            else:
                # in this case, the plot on the bottom right is rotated
                axs[0, 0].scatter(
                    ref_value_list[0],
                    0,
                    label="true value",
                    zorder=10,
                    **reference_values_kwargs,
                    edgecolor="black",
                )
                axs[1, 1].scatter(
                    0,
                    ref_value_list[1],
                    label="true value",
                    zorder=10,
                    **reference_values_kwargs,
                    edgecolor="black",
                )

        if plot_priors:

            # add the prior-pdfs to the marginal subplots
            prm_names = problem.get_theta_names(tex=False, components=False)
            i = 0  # not included in for-header due to possible dim-jumps
            for prm_name in prm_names:
                # for multivariate priors, no priors are plotted
                if problem.parameters[prm_name].dim > 1:
                    i += problem.parameters[prm_name].dim
                    continue
                x = None
                if focus_on_posterior:
                    if (problem.n_latent_prms_dim == 2) and (i == 1):
                        # the plot on the bottom right is rotated
                        x_min, x_max = axs[i, i].get_ylim()
                    else:
                        x_min, x_max = axs[i, i].get_xlim()
                    x = np.linspace(x_min, x_max, 200)
                # the following code adds labels to the prior and posterior plot if they
                # are represented as lines
                if axs[i, i].lines:
                    posterior_handle = [axs[i, i].lines[0]]
                    posterior_label = ["posterior"]
                else:
                    # this is for the case, when the posterior is not shown as a line,
                    # but for example as a histogram etc.
                    posterior_handle, posterior_label = [], []
                rotate = True if problem.n_latent_prms_dim == 2 and i == 1 else False
                problem.parameters[prm_name].prior.plot(
                    axs[i, i], problem.parameters, x=x, rotate=rotate
                )
                if show_legends:
                    prior_handle, prior_label = axs[i, i].get_legend_handles_labels()
                    axs[i, i].legend(
                        posterior_handle + prior_handle,
                        posterior_label + prior_label,
                        loc="upper right",
                    )
                i += 1

            # here, the axis of the non-marginal plots are adjusted to the new ranges
            if (not focus_on_posterior) and (problem.n_latent_prms_dim > 2):
                n = problem.n_latent_prms_dim
                for i in range(n):
                    x_min, x_max = axs[i, i].get_xlim()
                    for j in range(i + 1, n):
                        axs[j, i].set_xlim((x_min, x_max))
                    for j in range(0, i):
                        axs[i, j].set_ylim((x_min, x_max))
        else:

            # the following code adds legends to the marginal plots for the case where
            # no priors are supposed to be plotted
            if show_legends:
                prm_names = problem.get_theta_names(tex=False, components=True)
                for i, prm_name in enumerate(prm_names):
                    existing_handles, existing_labels = axs[
                        i, i
                    ].get_legend_handles_labels()
                    if axs[i, i].lines:
                        posterior_handle = [axs[i, i].lines[0]]
                        posterior_label = ["posterior"]
                    else:
                        # this is for the case, when the posterior is not shown as a
                        # line, but for example as a histogram etc.
                        posterior_handle, posterior_label = [], []
                    axs[i, i].legend(
                        posterior_handle + existing_handles,
                        posterior_label + existing_labels,
                        loc="upper right",
                    )

        # add a title to the plot, if requested
        if title:
            fig = plt.gcf()
            fig.suptitle(title, fontsize=title_size)

        # the following command reduces the otherwise wide margins; when only two
        # parameter (components) are given, the tight_layout()-call only results in a
        # warning without having an effect - hence, the if-clause
        if problem.n_latent_prms_dim > 2:
            plt.tight_layout()

        # by default, the y-axis of the first and last marginal plot have ticks, tick-
        # labels and axis-labels that are not meaningful to show on the y-axis; hence,
        # we remove them here; since the default plot looks different for only two
        # latent parameters, there is a check before
        if problem.n_latent_prms_dim > 2:
            for i in [0, -1]:
                axs[i, i].yaxis.set_ticks_position("none")
                axs[i, i].yaxis.set_ticklabels([])
                axs[i, i].yaxis.set_visible(False)
            for i in range(problem.n_latent_prms_dim - 1):
                axs[i, i].set_xticks(ticks=axs[-1, i].get_xticks())
                axs[i, i].set_xlim(axs[-1, i].get_xlim())
            axs[-1, -1].set_xticks(ticks=axs[-1, 0].get_yticks())
            axs[-1, -1].set_xlim(axs[-1, 0].get_ylim())

        return axs

    elif plot_with == "seaborn":
        raise NotImplementedError(
            "The plot-creation with seaborn has not been implemented yet."
        )

    elif plot_with == "matplotlib":
        raise NotImplementedError(
            "The plot-creation with matplotlib has not been implemented yet."
        )

    else:
        raise RuntimeError(
            f"Invalid 'plot_with' argument: '{plot_with}'. Available options are "
            f"currently 'arviz', 'seaborn', 'matplotlib'"
        )


def create_posterior_plot(
    inference_data: az.data.inference_data.InferenceData,
    problem: "InferenceProblem",
    plot_with: str = "arviz",
    kind: str = "hist",
    figsize: Optional[tuple] = (10, 3),
    textsize: Union[int, float] = 10,
    title_size: Union[int, float] = 14,
    title: Optional[str] = None,
    hdi_prob: float = 0.95,
    true_values: Optional[dict] = None,
    show: bool = True,
    **kwargs,
):
    """
    Creates a posterior-plot for the given inference data.

    Parameters
    ----------
    inference_data
        Contains the results of the sampling procedure.
    problem
        The inference problem the inference data refers to.
    plot_with
        Defines the python package the plot will be generated with. Options are:
        {'arviz', 'seaborn', 'matplotlib'}.
    kind
        Type of plot to display ('kde' or 'hist').
    figsize
        Defines the size of the generated plot in the default unit. If None is chosen,
        the figsize will be derived automatically.
    textsize
        Defines the font size in the default unit.
    title_size
        Defines the font size of the figures title if 'title' is given.
    title
        The title of the figure.
    hdi_prob
        Defines the highest density interval. Must be a number between 0 and 1.
    true_values
        Used for plotting 'true' parameter values. Keys are the parameter names and
        values are the values that are supposed to be shown in the marginal plots.
    show
        When True, the show-method is called after creating the plot. Otherwise, the
        show-method is not called. The latter is useful, when the plot should be further
        processed.
    kwargs
        Additional keyword arguments passed to arviz' plot_posterior function.

    Returns
    -------
    axs
        The array of subplots of the created plot.
    """

    if plot_with == "arviz":

        # process true_values if specified
        if true_values is not None:
            var_names_raw = problem.get_theta_names(tex=False)
            ref_val = []
            for var_name in var_names_raw:
                if len_or_one(true_values[var_name]) == 1:
                    ref_val.append(true_values[var_name])
                else:
                    for true_value in true_values[var_name]:
                        ref_val.append(true_value)
            kwargs["ref_val"] = ref_val

        # call the main plotting routine from arviz and return the axes object
        axs = az.plot_posterior(
            inference_data,
            kind=kind,
            figsize=figsize,
            textsize=textsize,
            hdi_prob=hdi_prob,
            show=show,
            **kwargs,
        )

        # add a title to the plot, if requested
        if title:
            fig = plt.gcf()
            fig.suptitle(title, fontsize=title_size)

        return axs

    elif plot_with == "seaborn":
        raise NotImplementedError(
            "The plot-creation with seaborn has not been implemented yet."
        )

    elif plot_with == "matplotlib":
        raise NotImplementedError(
            "The plot-creation with matplotlib has not been implemented yet."
        )

    else:
        raise RuntimeError(
            f"Invalid 'plot_with' argument: '{plot_with}'. Available options are "
            f"currently 'arviz', 'seaborn', 'matplotlib'"
        )


def create_trace_plot(
    inference_data: az.data.inference_data.InferenceData,
    problem: "InferenceProblem",
    plot_with: str = "arviz",
    kind: str = "trace",
    figsize: Optional[tuple] = (10, 6),
    textsize: Union[int, float] = 10,
    title_size: Union[int, float] = 14,
    title: Optional[str] = None,
    show: bool = True,
    **kwargs,
):
    """
    Creates a trace-plot for the given inference data.

    Parameters
    ----------
    inference_data
        Contains the results of the sampling procedure.
    problem
        The inference problem the inference data refers to.
    plot_with
        Defines the python package the plot will be generated with. Options are:
        {'arviz', 'seaborn', 'matplotlib'}.
    kind
        Allows to choose between plotting sampled values per iteration ("trace") and
        rank plots ("rank_bar", "rank_vlines").
    figsize
        Defines the size of the generated plot in the default unit. If None is chose,
        the figsize will be derived automatically.
    textsize
        Defines the font size in the default unit.
    title_size
        Defines the font size of the figures title if 'title' is given.
    title
        The title of the figure.
    show
        When True, the show-method is called after creating the plot. Otherwise, the
        show-method is not called. The latter is useful, when the plot should be further
        processed.
    kwargs
        Additional keyword arguments passed to arviz' plot_trace function.

    Returns
    -------
    axs
        The array of subplots of the created plot.
    """

    if plot_with == "arviz":

        # set default value for kde_kwargs if not given in kwargs; note that this
        # default value is mutable, so it should not be given as a default argument in
        # create_pair_plot
        if "plot_kwargs" not in kwargs:
            kwargs["plot_kwargs"] = {"textsize": textsize}

        # call the main plotting routine from arviz and return the axes object
        axs = az.plot_trace(
            inference_data, kind=kind, figsize=figsize, show=show, **kwargs
        )

        # add a title to the plot, if requested
        if title:
            fig = plt.gcf()
            fig.suptitle(title, fontsize=title_size)

        return axs

    elif plot_with == "seaborn":
        raise NotImplementedError(
            "The plot-creation with seaborn has not been implemented yet."
        )

    elif plot_with == "matplotlib":
        raise NotImplementedError(
            "The plot-creation with matplotlib has not been implemented yet."
        )

    else:
        raise RuntimeError(
            f"Invalid 'plot_with' argument: '{plot_with}'. Available options are "
            f"currently 'arviz', 'seaborn', 'matplotlib'"
        )



# standard library imports
import unittest

# third party imports
from scipy import stats

# local imports
from probeye.inference.scipy_.priors import PriorNormal
from probeye.inference.scipy_.priors import PriorTruncnormal
from probeye.inference.scipy_.priors import PriorLognormal
from probeye.inference.scipy_.priors import PriorUniform
from probeye.inference.scipy_.priors import PriorWeibull
from probeye.inference.scipy_.priors import translate_prior


class TestProblem(unittest.TestCase):
    def test_prior_normal(self):
        prior_normal = PriorNormal("a", ["loc_a", "scale_a"], "a_normal")
        # check the evaluation of the log-pdf
        prms = {"a": 1.0, "loc_a": 0.0, "scale_a": 1.0}
        self.assertEqual(
            stats.norm.logpdf(prms["a"], prms["loc_a"], prms["scale_a"]),
            prior_normal(prms, "logpdf"),
        )
        # check the sampling-method (samples are checked one by one)
        prms = {"loc_a": 0.0, "scale_a": 1.0}
        prior_samples = prior_normal.generate_samples(prms, 10, seed=1)
        sp_samples = stats.norm.rvs(
            loc=prms["loc_a"], scale=prms["scale_a"], size=10, random_state=1
        )
        for s1, s2 in zip(prior_samples, sp_samples):
            self.assertEqual(s1, s2)
        # test multivariate version
        prior_normal = PriorNormal("a", ["loc_a", "scale_a"], "a_normal")
        prms = {"loc_a": [0.0, 0.0], "scale_a": [1.0, 1.0]}
        sample = prior_normal(prms, method="rvs", use_ref_prm=False, size=10)
        self.assertEqual(len(sample), 10)
        # test requesting an invalid method
        with self.assertRaises(AttributeError):
            prior_normal(prms, method="invalid method")

    def test_prior_truncnormal(self):
        prior_truncnormal = PriorTruncnormal(
            "sigma", ["loc_sigma", "scale_sigma"], "sigma_normal"
        )
        # check the evaluation of the log-pdf
        prms = {
            "sigma": 1.0,
            "loc_sigma": 0.0,
            "scale_sigma": 1.0,
            "a_sigma": 0.0,
            "b_sigma": 5.0,
        }
        self.assertEqual(
            stats.truncnorm.logpdf(
                prms["sigma"],
                a=prms["a_sigma"],
                b=prms["b_sigma"],
                loc=prms["loc_sigma"],
                scale=prms["scale_sigma"],
            ),
            prior_truncnormal(prms, "logpdf"),
        )
        # check the sampling-method (samples are checked one by one)
        prms = {"loc_sigma": 0.0, "scale_sigma": 1.0, "a_sigma": 0.0, "b_sigma": 5.0}
        prior_samples = prior_truncnormal.generate_samples(prms, 10, seed=1)
        sp_samples = stats.truncnorm.rvs(
            a=prms["a_sigma"],
            b=prms["b_sigma"],
            loc=prms["loc_sigma"],
            scale=prms["scale_sigma"],
            size=10,
            random_state=1,
        )
        for s1, s2 in zip(prior_samples, sp_samples):
            self.assertEqual(s1, s2)

    def test_prior_lognormal(self):
        prior_lognormal = PriorLognormal("a", ["loc_a", "scale_a"], "a_lognormal")
        # check the evaluation of the log-pdf
        prms = {"a": 2.0, "loc_a": 1.0, "scale_a": 1.0}
        self.assertEqual(
            stats.lognorm.logpdf(prms["a"], prms["loc_a"], prms["scale_a"]),
            prior_lognormal(prms, "logpdf"),
        )
        # check the evaluation of the mean
        mean = prior_lognormal(prms, method="mean", use_ref_prm=False)
        self.assertAlmostEqual(mean, stats.lognorm.mean(1, loc=1, scale=1))
        # check the sampling-method (samples are checked one by one)
        prms = {"loc_a": 1.0, "scale_a": 1.0}
        prior_samples = prior_lognormal.generate_samples(prms, 10, seed=1)
        sp_samples = stats.lognorm.rvs(
            1.0,  # this is scipy's shape parameter
            loc=prms["loc_a"],
            scale=prms["scale_a"],
            size=10,
            random_state=1,
        )
        for s1, s2 in zip(prior_samples, sp_samples):
            self.assertEqual(s1, s2)

    def test_prior_uniform(self):
        prior_uniform = PriorUniform("a", ["low_a", "high_a"], "a_uniform")
        # check the evaluation of the log-pdf
        prms = {"a": 0.5, "low_a": 0.0, "high_a": 1.0}
        self.assertEqual(
            stats.uniform.logpdf(prms["a"], prms["low_a"], prms["high_a"]),
            prior_uniform(prms, "logpdf"),
        )
        # check the sampling-method (samples are checked one by one)
        prms = {"low_a": 0.0, "high_a": 1.0}
        prior_samples = prior_uniform.generate_samples(prms, 10, seed=1)
        sp_samples = stats.uniform.rvs(
            loc=prms["low_a"],
            scale=prms["low_a"] + prms["high_a"],
            size=10,
            random_state=1,
        )
        for s1, s2 in zip(prior_samples, sp_samples):
            self.assertEqual(s1, s2)

    def test_prior_weibull(self):
        prior_weibull = PriorWeibull("a", ["loc_a", "scale_a", "shape_a"], "a_weibull")
        # check the evaluation of the log-pdf
        prms = {"a": 1.0, "loc_a": 1.0, "scale_a": 1.0, "shape_a": 2.0}
        self.assertEqual(
            stats.weibull_min.logpdf(
                prms["a"], prms["shape_a"], prms["loc_a"], prms["scale_a"]
            ),
            prior_weibull(prms, "logpdf"),
        )
        # check the sampling-method (samples are checked one by one)
        prms = {"loc_a": 1.0, "scale_a": 1.0, "shape_a": 2.0}
        prior_samples = prior_weibull.generate_samples(prms, 10, seed=1)
        sp_samples = stats.weibull_min.rvs(
            prms["shape_a"],
            loc=prms["loc_a"],
            scale=prms["scale_a"],
            size=10,
            random_state=1,
        )
        for s1, s2 in zip(prior_samples, sp_samples):
            self.assertEqual(s1, s2)
        # check the evaluation of the mean
        mean = prior_weibull(prms, method="mean", use_ref_prm=False)
        self.assertAlmostEqual(mean, stats.weibull_min.mean(2, loc=1, scale=1))

    def test_translate_prior(self):
        # check the invalid prior_classes type
        prior_uniform = PriorUniform("a", ["low_a", "high_a"], "a_uniform")
        with self.assertRaises(TypeError):
            translate_prior(prior_uniform, prior_classes=[1, 2, 3])


if __name__ == "__main__":
    unittest.main()



# standard library imports
from typing import TYPE_CHECKING
import time
import random
import contextlib

# third party imports
import numpy as np
import dynesty
import arviz as az
from loguru import logger
from tabulate import tabulate
from scipy.stats import norm

# local imports
from probeye.subroutines import pretty_time_delta
from probeye.subroutines import check_for_uninformative_priors
from probeye.inference.scipy_.solver import ScipySolver
from probeye.subroutines import stream_to_logger, print_dict_in_rows

# imports only needed for type hints
if TYPE_CHECKING:  # pragma: no cover
    from probeye.definition.inference_problem import InferenceProblem


class DynestySolver(ScipySolver):
    """
    A static and dynamic nested parameter estimator. It facilitates the use of the
    python package `dynesty`. The default is set to a static parameter estimator.

    _Note:_ For full details on the `dynesty` library see:
        https://dynesty.readthedocs.io/en/latest/index.html.
    """

    def __init__(
        self, problem: "InferenceProblem", seed: int = 1, show_progress: bool = True
    ):
        """See docstring of ScipySolver for information on the arguments."""
        logger.debug("Initializing EmceeSolver")
        # check that the problem does not contain a uninformative prior
        check_for_uninformative_priors(problem)
        # initialize the scipy-based solver (ScipySolver)
        super().__init__(problem, seed=seed, show_progress=show_progress)

    def prior_transform(self, theta: np.ndarray) -> list:
        """
        Evaluates the ppf of the prior distributions at theta.

        Parameters
        ----------
        theta
            A numeric vector for which the ppf should be evaluated.
            Which parameters these numbers refer to can be checked by calling self.
            theta_explanation() once the problem is set up.

        Returns
        -------
        qs
            The vector of quantiles for each prior distribution at theta.
        """
        qs = []
        for prior in self.priors.values():
            prms = self.problem.get_parameters(theta, prior.prms_def)
            try:
                qs.append(prior(prms, "ppf"))
            except AttributeError as e:
                # This branch is active when there is no `ppf` method in
                # the prior distribution. For the case of a multivariate
                # normal distribution, we implement a workaround.
                loc = prms[f"loc_{prior.ref_prm}"]
                scale = prms[f"scale_{prior.ref_prm}"]
                x = prms[prior.ref_prm]

                # Assume the MVN to be uncorrelated, only diagonal nonzero
                i, j = np.nonzero(scale)
                assert np.all(i == j)

                mvn_qs = norm.ppf(q=x, loc=loc, scale=np.diagonal(scale))
                qs += list(mvn_qs)

        return qs

    def get_summary(self, posterior_samples: np.ndarray) -> dict:
        """
        Computes and prints a summary of the posterior samples containing mean, median,
        standard deviation, 5th percentile and 95th percentile. Note, that this method
        was based on code from the taralli package: https://gitlab.com/tno-bim/taralli.

        Parameters
        ----------
        posterior_samples
            The generated samples in an array with as many columns as there are latent
            parameters, and n rows, where n = n_chains * n_steps.

        Returns
        -------
            Keys are the different statistics 'mean', 'median', 'sd' (standard
            deviation), 'q05' and 'q95' (0.05- and 0.95-quantile). The values are
            dictionaries with the parameter names as keys and the respective statistics
            as values.
        """

        # used for the names in the first column
        var_names = self.problem.get_theta_names(tex=False, components=True)

        # compute some stats for each column (i.e., each parameter)
        mean = np.mean(posterior_samples, axis=0)
        quantiles = np.quantile(posterior_samples, [0.50, 0.05, 0.95], axis=0)
        median = quantiles[0, :]
        quantile_05 = quantiles[1, :]
        quantile_95 = quantiles[2, :]

        # compute the sample standard deviations for each parameter
        cov_matrix = np.atleast_2d(np.cov(posterior_samples.T))
        sd = np.sqrt(np.diag(cov_matrix))

        # assemble the summary array
        col_names = ["", "mean", "median", "sd", "5%", "95%"]
        row_names = np.array(var_names)
        tab = np.hstack(
            (
                row_names.reshape(-1, 1),
                mean.reshape(-1, 1),
                median.reshape(-1, 1),
                sd.reshape(-1, 1),
                quantile_05.reshape(-1, 1),
                quantile_95.reshape(-1, 1),
            )
        )

        # print the generated table, and return a summary dict for later use
        print(tabulate(tab, headers=col_names, floatfmt=".2f"))
        return {
            "mean": {name: val for name, val in zip(row_names, mean)},
            "median": {name: val for name, val in zip(row_names, median)},
            "sd": {name: val for name, val in zip(row_names, sd)},
            "q05": {name: val for name, val in zip(row_names, quantile_05)},
            "q95": {name: val for name, val in zip(row_names, quantile_95)},
        }

    def run_dynesty(
        self,
        estimation_method: str = "dynamic",
        nlive: int = 250,
        **kwargs,
    ) -> az.data.inference_data.InferenceData:
        """
        Runs the dynesty-sampler for the InferenceProblem the DynestySolver was
        initialized with and returns the results as an arviz InferenceData obj.

        Parameters
        ----------
        estimation_method
            "dynamic" or "static"
        nlive
            number of live points
        kwargs
            Additional key-word arguments channeled to emcee.EnsembleSampler.

        Returns
        -------
        inference_data or dynesty sampler
            Contains the results of the sampling procedure.
        """
        start = time.time()

        logger.info(f"Solving problem using dynesty sampler with {kwargs}")

        # The following code is based on taralli and merely adjusted to the variables
        # in the probeye setup; see https://gitlab.com/tno-bim/taralli

        # ............................................................................ #
        #                                 Pre-process                                  #
        # ............................................................................ #
        rstate = np.random.RandomState(self.seed)

        if estimation_method == "dynamic":
            sampler = dynesty.DynamicNestedSampler(
                loglikelihood=self.loglike,
                prior_transform=self.prior_transform,
                ndim=self.problem.n_latent_prms_dim,
                rstate=rstate,
                nlive=nlive,
            )
            sampler.run_nested(print_progress=self.show_progress, **kwargs)
        elif estimation_method == "static":
            sampler = dynesty.NestedSampler(
                loglikelihood=self.loglike,
                prior_transform=self.prior_transform,
                ndim=self.problem.n_latent_prms_dim,
                rstate=rstate,
                nlive=nlive,
            )
            sampler.run_nested(print_progress=self.show_progress, **kwargs)
        else:
            raise RuntimeError(
                "Choose 'dynamic' or 'static' as the estimation_method parameter!"
            )
        end = time.time()
        runtime_str = pretty_time_delta(end - start)
        logger.info(f"Total run-time: {runtime_str}.")
        logger.info("")

        logger.info("Resample weighted samples to equal samples for post")
        logger.info("processing. Access the original dynesty results via  .raw_results")
        weights = np.exp(sampler.results.logwt - sampler.results.logz[-1])
        samples = dynesty.utils.resample_equal(sampler.results.samples, weights)

        logger.info("Summary of sampling results")
        with contextlib.redirect_stdout(stream_to_logger("INFO")):  # type: ignore
            self.summary = self.get_summary(samples)
        logger.info(f"Posterior log evidence: {sampler.results.logz[-1]}")

        var_names = self.problem.get_theta_names(tex=True, components=True)

        posterior_dict = {name: samples[:, i] for i, name in enumerate(var_names)}

        inference_data = az.convert_to_inference_data(posterior_dict)
        # TODO maybe add information like "source = dynesty", timestamp, etc ...

        self.raw_results = sampler.results

        return inference_data



# standard library
from typing import Union, List, Optional

# third party imports
import numpy as np
from scipy import stats

# local imports
from probeye.definition.prior import PriorBase
from probeye.subroutines import len_or_one


class PriorNormal(PriorBase):
    """Prior class for a uni- or multivariate normal distribution."""

    def __init__(
        self,
        ref_prm: str,
        prms_def: Union[str, dict, List[Union[str, dict]]],
        name: str,
    ):
        """
        Parameters
        ----------
        ref_prm
            The name of the latent parameter the prior refers to.
        prms_def
            Defines the prior's parameter names. See the docstring of PriorBase for
            more detailed information.
        name
            Defining the priors name.
        """
        super().__init__(ref_prm, prms_def, name, "normal distribution")

    def __call__(
        self, prms: dict, method: str, use_ref_prm: bool = True, **kwargs
    ) -> float:
        """
        Evaluates stats.(multivariate_)norm.<method>(x, loc, scale) or, if use_ref_prm=
        False stats.(multivariate_)norm.<method>(loc, scale). This function is mostly
        used with method='logpdf' during the sampling procedure.

        Parameters
        ----------
        prms
            Contains the prior's parameters as keys and their values as values.
        method
            The method of stats.norm to be evaluated (e.g. 'pdf' or 'logpdf').
        use_ref_prm
            If True stats.(multivariate_)norm.<method>(x, loc, scale) is evaluated,
            hence 'x' must be provided in the prms dictionary. Otherwise, the evaluated
            method is stats.norm.<method>(loc, scale).
        kwargs
            Additional keyword arguments to pass to the specified method.

        Returns
        -------
            The result of stats.(multivariate_)norm.<method>(x, loc, scale) or of
            stats.(multivariate_)norm.<method>(loc, scale).
        """
        loc = prms[f"loc_{self.ref_prm}"]
        scale = prms[f"scale_{self.ref_prm}"]
        if len_or_one(loc) == 1:
            fun = getattr(stats.norm, method)
            if use_ref_prm:
                x = prms[self.ref_prm]
                return fun(x, loc=loc, scale=scale, **kwargs)
            else:
                return fun(loc=loc, scale=scale, **kwargs)
        else:
            try:
                fun = getattr(stats.multivariate_normal, method)
            except AttributeError:
                # this try-catch construct accounts for the fact, that the multivariate
                # normal distribution does not have a 'mean' or 'median' method
                if method in ["mean", "median"]:
                    return loc
                else:
                    raise AttributeError(
                        f"stats.multivariate_normal does "
                        f"not have a '{method}'-method"
                    )
            if use_ref_prm:
                x = prms[self.ref_prm]
                return fun(x, mean=loc, cov=scale, **kwargs)
            else:
                return fun(mean=loc, cov=scale, **kwargs)

    def generate_samples(
        self, prms: dict, size: int, seed: Optional[int] = None
    ) -> np.ndarray:
        """
        Randomly draws samples from this prior distribution. This method is used to
        create initial samples for MCMC-based algorithms.

        Parameters
        ----------
        prms
            Contains the prior's parameters as keys and their values as values.
        size
            Number of samples to generate.
        seed
            Used for the random state of the random number generation.

        Returns
        -------
        numpy.ndarray
            The generate samples.
        """
        loc = prms[f"loc_{self.ref_prm}"]
        scale = prms[f"scale_{self.ref_prm}"]
        if len_or_one(loc) == 1:
            return stats.norm.rvs(loc=loc, scale=scale, size=size, random_state=seed)
        else:
            return stats.multivariate_normal.rvs(
                mean=loc, cov=scale, size=size, random_state=seed
            )


class PriorLognormal(PriorBase):
    """Prior class for a log-normal distribution."""

    def __init__(
        self,
        ref_prm: str,
        prms_def: Union[str, dict, List[Union[str, dict]]],
        name: str,
    ):
        """
        Parameters
        ----------
        ref_prm
            The name of the latent parameter the prior refers to.
        prms_def
            Defines the prior's parameter names. See the docstring of PriorBase for
            more detailed information.
        name
            Defining the priors name.
        """
        super().__init__(ref_prm, prms_def, name, "log-normal distribution")

    def __call__(
        self,
        prms: dict,
        method: str,
        shape: Union[float, int] = 1,
        use_ref_prm: bool = True,
        **kwargs,
    ) -> float:
        """
        Evaluates stats.lognorm.<method>(x, loc, scale) or, if use_ref_prm=False stats.
        lognorm.<method>(loc, scale). This function is mostly used with method='logpdf'
        during the sampling procedure.

        Parameters
        ----------
        prms
            Contains the prior's parameters as keys and their values as values.
        method
            The method of stats.lognorm to be evaluated (e.g. 'pdf', 'logpdf').
        shape
            Scipy uses this shape parameter, which is not considered as a prior
            parameter here. So, it is set to 1, which results in the standard version
            of the lognormal distribution.
        use_ref_prm
            If True stats.lognorm.<method>(x, loc, scale) is evaluated, hence 'x' must
            be provided in the prms dictionary. Otherwise, the evaluated method is
            stats.lognorm.<method>(loc, scale).
        kwargs
            Additional keyword arguments to pass to the specified method.

        Returns
        -------
            The result of stats.lognorm.<method>(x, loc, scale) or of stats.lognorm.
            <method>(loc, scale).
        """
        fun = getattr(stats.lognorm, method)
        loc = prms[f"loc_{self.ref_prm}"]
        scale = prms[f"scale_{self.ref_prm}"]
        if use_ref_prm:
            x = prms[self.ref_prm]
            return fun(x, shape, loc=loc, scale=scale, **kwargs)
        else:
            return fun(shape, loc=loc, scale=scale, **kwargs)

    def generate_samples(
        self,
        prms: dict,
        size: int,
        seed: Optional[int] = None,
        shape: Union[float, int] = 1,
    ) -> np.ndarray:
        """
        Randomly draws samples from this prior distribution. This method is used to
        create initial samples for MCMC-based algorithms.

        Parameters
        ----------
        prms
            Contains the prior's parameters as keys and their values as values.
        size
            Number of samples to generate.
        seed
            Used for the random state of the random number generation.
        shape
            Scipy uses a shape parameter. For the common lognormal distribution this
            shape parameter is one.

        Returns
        -------
            The generate samples.
        """
        loc = prms[f"loc_{self.ref_prm}"]
        scale = prms[f"scale_{self.ref_prm}"]

        return stats.lognorm.rvs(
            shape, loc=loc, scale=scale, size=size, random_state=seed
        )


class PriorTruncnormal(PriorBase):
    """Prior class for a trunc-normal distribution."""

    def __init__(
        self,
        ref_prm: str,
        prms_def: Union[str, dict, List[Union[str, dict]]],
        name: str,
    ):
        """
        Parameters
        ----------
        ref_prm
            The name of the latent parameter the prior refers to.
        prms_def
            Defines the prior's parameter names. See the docstring of PriorBase for
            more detailed information.
        name
            Defining the priors name.
        """
        super().__init__(ref_prm, prms_def, name, "trunc-normal distribution")

    def __call__(
        self,
        prms: dict,
        method: str,
        use_ref_prm: bool = True,
        **kwargs,
    ) -> float:
        """
        Evaluates stats.truncnorm.<method>(x, a, b, loc, scale) or, if use_ref_prm=False
        stats.truncnorm.<method>(a, b, loc, scale). This function is mostly used with
        method='logpdf' during the sampling procedure.

        Parameters
        ----------
        prms
            Contains the prior's parameters as keys and their values as values.
        method
            The method of stats.truncnorm to be evaluated (e.g. 'pdf', 'logpdf').
        use_ref_prm
            If True stats.truncnorm.<method>(x, loc, scale) is evaluated, hence 'x' must
            be provided in the prms dictionary. Otherwise, the evaluated method is
            stats.truncnorm.<method>(loc, scale).
        kwargs
            Additional keyword arguments to pass to the specified method.

        Returns
        -------
            The result of stats.truncnorm.<method>(x, loc, scale) or of stats.truncnorm.
            <method>(loc, scale).
        """
        fun = getattr(stats.truncnorm, method)
        loc = prms[f"loc_{self.ref_prm}"]
        scale = prms[f"scale_{self.ref_prm}"]
        a = (prms[f"a_{self.ref_prm}"] - loc) / scale
        b = (prms[f"b_{self.ref_prm}"] - loc) / scale
        if use_ref_prm:
            x = prms[self.ref_prm]
            return fun(x, a=a, b=b, loc=loc, scale=scale, **kwargs)
        else:
            return fun(a=a, b=b, loc=loc, scale=scale, **kwargs)

    def generate_samples(
        self,
        prms: dict,
        size: int,
        seed: Optional[int] = None,
    ) -> np.ndarray:
        """
        Randomly draws samples from this prior distribution. This method is used to
        create initial samples for MCMC-based algorithms.

        Parameters
        ----------
        prms
            Contains the prior's parameters as keys and their values as values.
        size
            Number of samples to generate.
        seed
            Used for the random state of the random number generation.

        Returns
        -------
            The generate samples.
        """
        loc = prms[f"loc_{self.ref_prm}"]
        scale = prms[f"scale_{self.ref_prm}"]
        a = (prms[f"a_{self.ref_prm}"] - loc) / scale
        b = (prms[f"b_{self.ref_prm}"] - loc) / scale
        return stats.truncnorm.rvs(
            a=a, b=b, loc=loc, scale=scale, size=size, random_state=seed
        )


class PriorUniform(PriorBase):
    """Prior class for a uniform distribution."""

    def __init__(
        self,
        ref_prm: str,
        prms_def: Union[str, dict, List[Union[str, dict]]],
        name: str,
    ):
        """
        Parameters
        ----------
        ref_prm
            The name of the latent parameter the prior refers to.
        prms_def
            Defines the prior's parameter names. See the docstring of PriorBase for
            more detailed information.
        name
            Defining the priors name.
        """
        super().__init__(ref_prm, prms_def, name, "uniform distribution")

    def __call__(
        self, prms: dict, method: str, use_ref_prm: bool = True, **kwargs
    ) -> float:
        """
        Evaluates stats.uniform.<method>(x, loc, scale) or, if use_ref_prm=False stats.
        uniform.<method>(loc, scale). This function is mostly used with method='logpdf'
        during the sampling procedure.

        Parameters
        ----------
        prms
            Contains the prior's parameters as keys and their values as values.
        method
            The method of stats.uniform to be evaluated (e.g. 'pdf', 'logpdf').
        use_ref_prm
            If True stats.uniform.<method>(x, loc, scale) is evaluated, hence 'x' must
            be provided in the prms dictionary. Otherwise, the evaluated method is
            stats.uniform.<method>(loc, scale).
        kwargs
            Additional keyword arguments to pass to the specified method.

        Returns
        -------
            The result of stats.uniform.<method>(x, loc, scale) or of stats.uniform.
            <method>(loc, scale).
        """
        fun = getattr(stats.uniform, method)
        low = prms[f"low_{self.ref_prm}"]
        high = prms[f"high_{self.ref_prm}"]
        if use_ref_prm:
            x = prms[self.ref_prm]
            return fun(x, loc=low, scale=high - low, **kwargs)
        else:
            return fun(loc=low, scale=high - low, **kwargs)

    def generate_samples(
        self, prms: dict, size: int, seed: Optional[int] = None
    ) -> np.ndarray:
        """
        Randomly draws samples from this prior distribution. This method is used to
        create initial samples for MCMC-based algorithms.

        Parameters
        ----------
        prms
            Contains the prior's parameters as keys and their values as values.
        size
            Number of samples to generate.
        seed
            Used for the random state of the random number generation.

        Returns
        -------
            The generate samples.
        """
        low = prms[f"low_{self.ref_prm}"]
        high = prms[f"high_{self.ref_prm}"]
        return stats.uniform.rvs(
            loc=low, scale=high - low, size=size, random_state=seed
        )


class PriorWeibull(PriorBase):
    """Prior class for a three-parameter Weibull distribution."""

    def __init__(
        self,
        ref_prm: str,
        prms_def: Union[str, dict, List[Union[str, dict]]],
        name: str,
    ):
        """
        Parameters
        ----------
        ref_prm
            The name of the latent parameter the prior refers to.
        prms_def
            Defines the prior's parameter names. See the docstring of PriorBase for
            more detailed information.
        name
            Defining the priors name.
        """
        super().__init__(ref_prm, prms_def, name, "Weibull distribution")

    def __call__(
        self, prms: dict, method: str, use_ref_prm: bool = True, **kwargs
    ) -> float:
        """
        Evaluates stats.weibull_min.<method>(x, loc, scale) or, if use_ref_prm=False,
        stats.weibull_min.<method>(loc, scale). This function is mostly used with
        method='logpdf' during the sampling procedure.

        Parameters
        ----------
        prms
            Contains the prior's parameters as keys and their values as values.
        method
            The method of stats.weibull_min to be evaluated ('pdf', 'logpdf', etc.).
        use_ref_prm
            If True stats.weibull_min.<method>(x, loc, scale) is evaluated, hence 'x'
            must be provided in the prms dictionary. Otherwise, the evaluated method is
            weibull_min.<method>(loc, scale).
        kwargs
            Additional keyword arguments to pass to the specified method.

        Returns
        -------
            The result of stats.weibull_min.<method>(x, loc, scale) or of stats.
            weibull_min.<method>(loc, scale).
        """
        fun = getattr(stats.weibull_min, method)
        shape = prms[f"shape_{self.ref_prm}"]
        loc = prms[f"loc_{self.ref_prm}"]
        scale = prms[f"scale_{self.ref_prm}"]
        if use_ref_prm:
            x = prms[self.ref_prm]
            return fun(x, shape, loc=loc, scale=scale, **kwargs)
        else:
            return fun(shape, loc=loc, scale=scale, **kwargs)

    def generate_samples(
        self, prms: dict, size: int, seed: Optional[int] = None
    ) -> np.ndarray:
        """
        Randomly draws samples from this prior distribution. This method is used to
        create initial samples for MCMC-based algorithms.

        Parameters
        ----------
        prms
            Contains the prior's parameters as keys and their values as values.
        size
            Number of samples to generate.
        seed
            Used for the random state of the random number generation.

        Returns
        -------
            The generate samples.
        """
        shape = prms[f"shape_{self.ref_prm}"]
        loc = prms[f"loc_{self.ref_prm}"]
        scale = prms[f"scale_{self.ref_prm}"]
        return stats.weibull_min.rvs(
            shape, loc=loc, scale=scale, size=size, random_state=seed
        )


def translate_prior(
    prior_template: PriorBase, prior_classes: Optional[dict] = None
) -> PriorBase:
    """
    Translate a given instance of PriorBase (which is essentially just a description of
    the prior without compute-methods) to a specific prior object which does contain
    compute-methods (e.g. compute the logpdf).

    Parameters
    ----------
    prior_template
        An instance of PriorBase which contains basic information on the prior but no
        computing-methods.
    prior_classes
        If None, the prior classes of defined in this file are used to translate the
        prior_template. If other user-defined priors should be used, they can be
        provided via this argument, by providing a dictionary with the prior_type as
        key and the custom prior class as value.

    Returns
    -------
    prior_object
        An instance of a specific prior class with computing capabilities. Examples for
        such classes are given above in this file.
    """

    # no translation is required for an uninformative prior
    if prior_template.prior_type == "uninformative":
        return prior_template

    # check the prior_classes argument; it either must be None, or of type dict
    if type(prior_classes) is not dict:
        if prior_classes is None:
            prior_classes = {
                "normal": PriorNormal,
                "lognormal": PriorLognormal,
                "truncnormal": PriorTruncnormal,
                "uniform": PriorUniform,
                "weibull": PriorWeibull,
            }
        else:
            # in this case prior_classes is not None, and not of type dict
            raise TypeError(
                f"Custom prior_classes must be given as a dictionary. However, "
                f"you provided an input of type {type(prior_classes)}."
            )

    # prepare the corresponding prior object; the following translation is necessary,
    # because prms_def must be given in form of a list, but was already translated to
    # a dictionary when instantiating the PriorBase objects; hence prior_template.
    # prms_def is a dictionary
    prms_def = [
        {key: value}
        for key, value in prior_template.prms_def.items()
        if key != prior_template.ref_prm
    ]
    prior_object = prior_classes[prior_template.prior_type](
        prior_template.ref_prm, prms_def, prior_template.name
    )

    return prior_object



# standard library
import unittest
import io
import sys

# third party imports
import numpy as np

# local imports
from probeye.definition.inference_problem import InferenceProblem
from probeye.definition.forward_model import ForwardModelBase
from probeye.definition.sensor import Sensor
from probeye.definition.noise_model import NoiseModelBase


class TestProblem(unittest.TestCase):
    def test_properties(self):
        p = InferenceProblem("TestProblem")
        # add first latent parameter; n_latent_prms should be 1
        p.add_parameter("a", "model", prior=("normal", {"loc": 0.0, "scale": 1.0}))
        self.assertEqual(p.n_latent_prms, 1)
        self.assertEqual(p.latent_prms_dims, [1])
        # add second latent parameter; n_latent_prms should be 2
        p.add_parameter("b", "noise", prior=("normal", {"loc": 0.0, "scale": 1.0}))
        self.assertEqual(p.n_latent_prms, 2)
        self.assertEqual(p.latent_prms_dims, [1, 1])
        # check the different properties
        self.assertEqual(p.n_prms, 6)
        self.assertEqual(
            set(p.prms), {"loc_a", "scale_a", "a", "loc_b", "scale_b", "b"}
        )
        self.assertEqual(p.n_latent_prms, 2)
        self.assertEqual(set(p.latent_prms), {"a", "b"})
        self.assertEqual(p.n_constant_prms, 4)
        self.assertEqual(set(p.constant_prms), {"loc_a", "scale_a", "loc_b", "scale_b"})
        self.assertEqual(p.n_model_prms, 1)
        self.assertEqual(set(p.model_prms), {"a"})
        self.assertEqual(p.n_prior_prms, 4)
        self.assertEqual(set(p.prior_prms), {"loc_a", "scale_a", "loc_b", "scale_b"})
        self.assertEqual(p.n_noise_prms, 1)
        self.assertEqual(set(p.noise_prms), {"b"})

    def test_info(self):
        # simply check that no errors occur when info-method is called; the output that
        # is usually printed is redirected
        p = InferenceProblem("TestProblem")
        p.add_parameter("a", "model", const=1.0)
        p.add_parameter("b", "model", prior=("normal", {"loc": 0, "scale": 1}))
        p.add_parameter("s", "noise", prior=("normal", {"loc": 0, "scale": 1}))
        sys.stdout = io.StringIO()
        # try out different options
        p.info(include_experiments=False, tablefmt="presto", check_consistency=False)
        _ = p.info(
            include_experiments=False, tablefmt="presto", check_consistency=False
        )
        p.info(include_experiments=True, tablefmt="presto", check_consistency=False)
        p.info(include_experiments=True, tablefmt="plain", check_consistency=False)
        sys.stdout = sys.__stdout__  # reset printout to console
        with self.assertRaises(AssertionError):
            # the problem is not consistent yet (e.g. no forward model defined yet), so
            # the consistency_check will raise an error
            p.info(include_experiments=True, tablefmt="presto", check_consistency=True)
        # now add the remaining stuff to make to problem consistent
        test_model = ForwardModelBase("b", Sensor("x"), Sensor("y"))
        p.add_forward_model("TestModel", test_model)
        p.add_noise_model(NoiseModelBase("normal", "s", sensors=Sensor("y")))
        p.add_experiment(
            "Experiment_1", sensor_values={"x": 1, "y": 1}, fwd_model_name="TestModel"
        )
        sys.stdout = io.StringIO()
        # now, the consistency_check should not raise an error
        p.info(include_experiments=True, tablefmt="presto", check_consistency=True)
        sys.stdout = sys.__stdout__  # reset printout to console

    def test_str(self):
        # set up a consistent problem and print it
        p = InferenceProblem("TestProblem")
        p.add_parameter("a", "model", const=1.0)
        p.add_parameter("b", "model", prior=("normal", {"loc": 0, "scale": 1}))
        p.add_parameter("s", "noise", prior=("normal", {"loc": 0, "scale": 1}))
        test_model = ForwardModelBase("b", Sensor("x"), Sensor("y"))
        p.add_forward_model("TestModel", test_model)
        p.add_noise_model(NoiseModelBase("normal", "s", sensors=Sensor("y")))
        p.add_experiment(
            "Experiment_1", sensor_values={"x": 1, "y": 1}, fwd_model_name="TestModel"
        )
        sys.stdout = io.StringIO()  # redirect output to console
        print(p)
        sys.stdout = sys.__stdout__  # reset printout to console

    def test_add_parameter(self):
        p = InferenceProblem("TestProblem")
        # check valid use cases for constant parameters
        p.add_parameter("c", "model", const=1.0, info="info", tex=r"$c$")
        p.add_parameter("loc_a", "prior", const=1.0, info="info", tex=r"$loc_a$")
        p.add_parameter("sigma_1", "noise", const=1.0, info="info", tex=r"$\sigma_1$")
        # check valid use cases for latent parameters
        p.add_parameter(
            "b",
            "model",
            prior=("normal", {"loc": 0, "scale": 1}),
            info="info",
            tex="$b$",
        )
        p.add_parameter(
            "scale_a",
            "prior",
            info="info",
            tex="$b$",
            prior=("uniform", {"low": 0.1, "high": 2.0}),
        )
        p.add_parameter(
            "sigma_2",
            "noise",
            info="info",
            tex=r"$\sigma_2$",
            prior=("uniform", {"low": 0.1, "high": 2.0}),
        )
        p.add_parameter(
            "a",
            "model",
            info="info",
            tex="$a$",
            prior=("normal", {"loc": "loc_a", "scale": "scale_a"}),
        )
        p.add_parameter("d", "model")  # latent param. with uninformative prior
        # check invalid input arguments
        with self.assertRaises(RuntimeError):
            # adding a parameter with wrong type
            p.add_parameter("d", "wrong_type", const=1.0)
        with self.assertRaises(RuntimeError):
            # adding a parameter with both const and prior given
            p.add_parameter(
                "a", "model", const=1.0, prior=("uniform", {"low": 0.1, "high": 2.0})
            )
        with self.assertRaises(RuntimeError):
            # adding a parameter with a name that was used before
            p.add_parameter("c", "noise", const=2.0, info="info", tex=r"$c$")
        with self.assertRaises(TypeError):
            # adding a parameter with an invalid prior
            p.add_parameter(
                "eps",
                "model",
                info="info",
                tex=r"$\epsilon$",
                prior="Wrong prior format",
            )
        with self.assertRaises(RuntimeError):
            # adding a parameter with an invalid prior
            p.add_parameter(
                "eps",
                "model",
                info="info",
                tex=r"$\epsilon$",
                prior=[{"low": 0.1, "high": 2.0}],
            )
        with self.assertRaises(RuntimeError):
            # adding a parameter with an invalid prior
            p.add_parameter(
                "eps",
                "model",
                info="info",
                tex=r"$\epsilon$",
                prior=("uniform", {"low": 0.1, "high": 2.0}, "3"),
            )
        with self.assertRaises(TypeError):
            # adding a parameter with an invalid prior
            p.add_parameter(
                "eps",
                "model",
                info="info",
                tex=r"$\epsilon$",
                prior=(None, {"low": 0.1, "high": 2.0}),
            )
        with self.assertRaises(TypeError):
            # adding a parameter with an invalid prior
            p.add_parameter(
                "eps",
                "model",
                info="info",
                tex=r"$\epsilon$",
                prior=("uniform", [{"low": 0.1}, {"high": 2.0}]),
            )
        with self.assertRaises(TypeError):
            # adding a parameter with an invalid prior
            p.add_parameter(
                "eps",
                "model",
                info="info",
                tex=r"$\epsilon$",
                prior=("uniform", {"low": None, "high": 2.0}),
            )

    def test_remove_parameter(self):
        # set up a problem with some parameters so we can remove them
        p = InferenceProblem("TestProblem")
        p.add_parameter("c", "model", const=1.0, info="info", tex=r"$c$")
        p.add_parameter("loc_a", "prior", const=1.0, info="info", tex=r"$loc_a$")
        p.add_parameter("s", "noise", const=1.0, info="info", tex=r"$\sigma$")
        p.add_parameter(
            "scale_a",
            "prior",
            info="info",
            tex="$b$",
            prior=("uniform", {"low": 0.1, "high": 2.0}),
        )
        p.add_parameter(
            "sigma",
            "noise",
            info="info",
            tex=r"$\sigma$",
            prior=("uniform", {"low": 0.1, "high": 2.0}),
        )
        p.add_parameter(
            "a",
            "model",
            info="info",
            tex="$a$",
            prior=("normal", {"loc": "loc_a", "scale": "scale_a"}),
        )
        # check removing a constant parameter
        self.assertEqual(
            set(p.constant_prms),
            {
                "c",
                "loc_a",
                "s",
                "low_scale_a",
                "high_scale_a",
                "low_sigma",
                "high_sigma",
            },
        )
        p.remove_parameter("c")  # <-- this is where the removal happens
        self.assertEqual(
            set(p.constant_prms),
            {"loc_a", "s", "low_scale_a", "high_scale_a", "low_sigma", "high_sigma"},
        )
        # check removing a latent parameter; note that removing a latent parameter leads
        # to the removal of all its prior parameters
        self.assertEqual(set(p.latent_prms), {"scale_a", "sigma", "a"})
        # check that indexes are as expected
        self.assertEqual(p.parameters["scale_a"].index, 0)
        self.assertEqual(p.parameters["sigma"].index, 1)
        self.assertEqual(p.parameters["a"].index, 2)
        p.remove_parameter("a")  # <-- this is where the removal happens
        # check that all latent parameters associated with 'a' are gone
        self.assertEqual(set(p.latent_prms), {"sigma"})
        # check if the re-indexing worked
        self.assertEqual(p.parameters["sigma"].index, 0)
        # check that all constant prior-parameters associated with 'a' are gone
        self.assertEqual(set(p.constant_prms), {"s", "low_sigma", "high_sigma"})
        # check that the removal of non-existing parameters results in an error
        with self.assertRaises(RuntimeError):
            p.remove_parameter("undefined")

    def test_check_if_parameter_exists(self):
        # check confirming existing parameter
        p = InferenceProblem("TestProblem")
        p.add_parameter("a", "model", prior=("normal", {"loc": 0, "scale": 1}))
        p.parameters.confirm_that_parameter_exists("a")
        # check RuntimeError for non-existing parameter
        with self.assertRaises(RuntimeError):
            p.parameters.confirm_that_parameter_exists("b")

    def test_change_parameter_role(self):
        p = InferenceProblem("TestProblem")
        # check change of role from latent to constant parameter
        p.add_parameter("a", "model", prior=("normal", {"loc": 0, "scale": 1}))
        self.assertEqual(set(p.latent_prms), {"a"})
        self.assertEqual(set(p.constant_prms), {"loc_a", "scale_a"})
        self.assertEqual(set(p.prms), {"a", "loc_a", "scale_a"})
        p.change_parameter_role("a", const=1.0)  # <-- here the role changes
        with self.assertRaises(RuntimeError):
            # trying to change a constant to a constant
            p.change_parameter_role("a", const=1.0)
        self.assertEqual(set(p.latent_prms), set())
        self.assertEqual(set(p.constant_prms), {"a"})
        self.assertEqual(set(p.prms), {"a"})
        # check change of role from latent to constant parameter
        p.change_parameter_role("a", prior=("normal", {"loc": 0, "scale": 1}))
        self.assertEqual(set(p.latent_prms), {"a"})
        self.assertEqual(set(p.constant_prms), {"loc_a", "scale_a"})
        self.assertEqual(set(p.prms), {"a", "loc_a", "scale_a"})
        # check invalid input arguments
        with self.assertRaises(RuntimeError):
            # stated parameter does not exist
            p.change_parameter_role("undefined", const=1.0)
        with self.assertRaises(RuntimeError):
            # neither const nor prior are given
            p.change_parameter_role("a")
        with self.assertRaises(RuntimeError):
            # both const and prior are given
            p.change_parameter_role(
                "a", const=1.0, prior=("normal", {"loc": 0, "scale": 1})
            )
        with self.assertRaises(RuntimeError):
            # change to role the parameter already has
            p.change_parameter_role("a", prior=("normal", {"loc": 0, "scale": 1}))

    def test_change_parameter_info(self):
        p = InferenceProblem("TestProblem")
        # simple check that the change works and has the expected effect
        p.add_parameter(
            "a",
            "model",
            prior=("normal", {"loc": 0, "scale": 1}),
            info="Info",
            tex="$a$",
        )
        new_info, new_tex = "New info", r"$\hat{a}$"
        p.change_parameter_info("a", new_info, new_tex=new_tex)
        self.assertEqual(p.parameters["a"].info, new_info)
        self.assertEqual(p.parameters["a"].tex, new_tex)
        # check that you can just change either the info- or the tex-string
        new_info, new_tex = "Even newer info", r"$\hat{a}_{wow}$"
        p.change_parameter_info("a", new_info=new_info)
        p.change_parameter_info("a", new_tex=new_tex)
        self.assertEqual(p.parameters["a"].info, new_info)
        self.assertEqual(p.parameters["a"].tex, new_tex)
        # check call for non-existing parameter
        with self.assertRaises(RuntimeError):
            p.change_parameter_info("b", new_info, new_tex=new_tex)

    def test_change_constant(self):
        p = InferenceProblem("TestProblem")
        # simple check that the change works and has the expected effect
        p.add_parameter("a", "model", const=0.0)
        p.add_parameter("b", "model", prior=("normal", {"loc": 0, "scale": 1}))
        new_const = 1.0
        p.change_constant("a", new_const)
        self.assertEqual(p.parameters["a"].value, new_const)
        # check for invalid input arguments
        with self.assertRaises(RuntimeError):
            # given parameter does not exist
            p.change_constant("undefined", 2.0)
        with self.assertRaises(RuntimeError):
            # given parameter is not a constant parameter
            p.change_constant("b", 2.0)

    def test_check_problem_consistency(self):
        p = InferenceProblem("TestProblem")
        # check if the method identifies inconsistent problems
        with self.assertRaises(AssertionError):
            # nothing is defined at this point
            p.check_problem_consistency()
        # define some parameters that do not require priors
        p.add_parameter("loc_a", "prior", const=0.0)
        p.add_parameter("scale_a", "prior", const=1.0)
        with self.assertRaises(AssertionError):
            # no priors defined yet
            p.check_problem_consistency()
        # add a prior by defining a latent parameter
        p.add_parameter(
            "a",
            "model",
            info="Some model parameter",
            tex="$a$",
            prior=("normal", {"loc": "loc_a", "scale": "scale_a"}),
        )
        with self.assertRaises(AssertionError):
            # no forward models defined yet
            p.check_problem_consistency()
        # add a forward model
        test_model = ForwardModelBase("a", Sensor("x"), Sensor("y"))
        p.add_forward_model("TestModel", test_model)
        with self.assertRaises(AssertionError):
            # no noise models defined yet
            p.check_problem_consistency()
        # add a noise model
        p.add_parameter("s", "noise", const=1.0)
        noise_model = NoiseModelBase("normal", "s", sensors=[Sensor("y")])
        p.add_noise_model(noise_model)
        with self.assertRaises(AssertionError):
            # no experiment_names defined yet
            p.check_problem_consistency()
        # add at an experiment
        p.add_experiment(
            "Experiment_1", sensor_values={"x": 1, "y": 1}, fwd_model_name="TestModel"
        )
        # now the problem should be consistent
        p.check_problem_consistency()

    def test_add_experiment(self):
        # check correct use
        p = InferenceProblem("TestProblem")
        p.add_parameter("a", "model", prior=("normal", {"loc": 0, "scale": 1}))
        test_model = ForwardModelBase("a", Sensor("x"), Sensor("y"))
        p.add_forward_model("TestModel", test_model)
        p.add_experiment(
            "Experiment_1", sensor_values={"x": 1, "y": 1}, fwd_model_name="TestModel"
        )
        # the experiment may contain more sensors than needed by forward model
        p.add_experiment(
            "Experiment_2",
            sensor_values={"x": 1, "y": 1, "z": 1},
            fwd_model_name="TestModel",
        )
        # check that scalar sensor_values are not transformed to numpy arrays
        x_in_p = p.experiments["Experiment_2"]["sensor_values"]["x"]
        self.assertEqual(type(x_in_p), type(1))
        # check invalid input arguments
        with self.assertRaises(TypeError):
            # wrong sensor_values type
            # noinspection PyTypeChecker
            p.add_experiment(
                "Experiment_3",
                sensor_values=[("x", 1), ("y", 2)],
                fwd_model_name="TestModel",
            )
        with self.assertRaises(TypeError):
            # wrong fwd_model_name type
            # noinspection PyTypeChecker
            p.add_experiment(
                "Experiment_3", sensor_values={"x": 1, "y": 1}, fwd_model_name=1.2
            )
        with self.assertRaises(RuntimeError):
            # referencing non-existing forward model
            p.add_experiment(
                "Experiment_3",
                sensor_values={"x": 1, "y": 1},
                fwd_model_name="ThisModelDoesNotExist",
            )
        with self.assertRaises(RuntimeError):
            # forward model's input sensor not provided by experiment
            p.add_experiment(
                "Experiment_3", sensor_values={"y": 1}, fwd_model_name="TestModel"
            )
        with self.assertRaises(RuntimeError):
            # forward model's output sensor not provided by experiment
            p.add_experiment(
                "Experiment_3", sensor_values={"x": 1}, fwd_model_name="TestModel"
            )
        with self.assertRaises(RuntimeError):
            # sensor values with different lengths
            p.add_experiment(
                "Experiment_3",
                fwd_model_name="TestModel",
                sensor_values={"x": [1, 2], "y": 1},
            )
        # check that sensor_value lists are transformed to numpy arrays
        p.add_experiment(
            "Experiment_3",
            fwd_model_name="TestModel",
            sensor_values={"x": [1, 2], "y": [2, 3]},
        )
        x_in_p = p.experiments["Experiment_3"]["sensor_values"]["x"]
        y_in_p = p.experiments["Experiment_3"]["sensor_values"]["y"]
        self.assertEqual(type(x_in_p), np.ndarray)
        self.assertEqual(type(y_in_p), np.ndarray)
        # check adding the same experiment again
        p.add_experiment(
            "Experiment_1", sensor_values={"x": 1, "y": 1}, fwd_model_name="TestModel"
        )

    def test_get_parameters(self):
        # check a simple use case
        p = InferenceProblem("TestProblem")
        p.add_parameter("c", "model", const=1.0)
        p.add_parameter("a", "model", prior=("normal", {"loc": 0, "scale": 1}))
        p.add_parameter("b", "model", prior=("normal", {"loc": 0, "scale": 1}))
        a_value, b_value = 3.1, 14.7
        prms_def = {"b": "b", "a": "a", "loc_b": "loc_b", "c": "c"}
        computed_result = p.get_parameters([a_value, b_value], prms_def)
        expected_result = {"b": b_value, "a": a_value, "loc_b": 0, "c": 1.0}
        self.assertEqual(computed_result, expected_result)

    def test_get_experiment_names(self):
        # prepare for checks
        p = InferenceProblem("TestProblem")
        p.add_parameter("a", "model", prior=("normal", {"loc": 0, "scale": 1}))
        # define two forward models
        test_model_1 = ForwardModelBase("a", Sensor("x"), [Sensor("y1"), Sensor("y2")])
        test_model_2 = ForwardModelBase("a", Sensor("x"), [Sensor("z1"), Sensor("z2")])
        p.add_forward_model("TestModel_1", test_model_1)
        p.add_forward_model("TestModel_2", test_model_2)
        # define experiment_names to each forward model
        p.add_experiment(
            "Experiment_Y1",
            sensor_values={"x": 1, "y1": 1, "y2": 1},
            fwd_model_name="TestModel_1",
        )
        p.add_experiment(
            "Experiment_Y2",
            sensor_values={"x": 2, "y1": 2, "y2": 2},
            fwd_model_name="TestModel_1",
        )
        p.add_experiment(
            "Experiment_Z1",
            sensor_values={"x": -1, "z1": -1, "z2": -1},
            fwd_model_name="TestModel_2",
        )
        p.add_experiment(
            "Experiment_Z2",
            sensor_values={"x": -2, "z1": -2, "z2": -2},
            fwd_model_name="TestModel_2",
        )
        # get experiment_names for specific forward model
        computed_result = p.get_experiment_names(forward_model_names="TestModel_1")
        expected_result = ["Experiment_Y1", "Experiment_Y2"]
        for i in range(len(computed_result)):
            self.assertEqual(computed_result[i], expected_result[i])
        # get experiment_names for specific forward model
        computed_result = p.get_experiment_names(forward_model_names="TestModel_2")
        expected_result = ["Experiment_Z1", "Experiment_Z2"]
        for i in range(len(computed_result)):
            self.assertEqual(computed_result[i], expected_result[i])
        # get experiment_names for both forward models
        computed_result = p.get_experiment_names(
            forward_model_names=["TestModel_1", "TestModel_2"]
        )
        expected_result = [
            "Experiment_Y1",
            "Experiment_Y2",
            "Experiment_Z1",
            "Experiment_Z2",
        ]
        for i in range(len(computed_result)):
            self.assertEqual(computed_result[i], expected_result[i])
        # get experiment_names via output sensors
        computed_result = p.get_experiment_names(sensor_names=["y1", "y2"])
        expected_result = ["Experiment_Y1", "Experiment_Y2"]
        for i in range(len(computed_result)):
            self.assertEqual(computed_result[i], expected_result[i])
        # get experiment_names via output sensors
        computed_result = p.get_experiment_names(sensor_names=["z1", "z2"])
        expected_result = ["Experiment_Z1", "Experiment_Z2"]
        for i in range(len(computed_result)):
            self.assertEqual(computed_result[i], expected_result[i])
        # get experiment_names via output sensors
        computed_result = p.get_experiment_names(sensor_names=["z1", "y1", "z2"])
        expected_result = []
        self.assertEqual(computed_result, expected_result)
        # get experiment_names via both forward model name and sensors
        computed_result = p.get_experiment_names(
            sensor_names=["y1", "y2"], forward_model_names="TestModel_1"
        )
        expected_result = ["Experiment_Y1", "Experiment_Y2"]
        for i in range(len(computed_result)):
            self.assertEqual(computed_result[i], expected_result[i])
        # get experiment_names via both forward model name and sensors
        computed_result = p.get_experiment_names(
            sensor_names=["y1", "y2"], forward_model_names="TestModel_2"
        )
        expected_result = []
        self.assertEqual(computed_result, expected_result)
        # check invalid input arguments
        with self.assertRaises(RuntimeError):
            # neither forward_model_names nor sensor_names given
            p.get_experiment_names()

    def test_get_theta_names(self):
        # check some simple use cases
        p = InferenceProblem("TestProblem")
        p.add_parameter(
            "a", "model", prior=("normal", {"loc": 0, "scale": 1}), tex="$a$"
        )
        p.add_parameter(
            "b", "model", prior=("normal", {"loc": 0, "scale": 1}), tex="$b$"
        )
        computed_result = p.get_theta_names(tex=False)
        expected_result = ["a", "b"]
        self.assertEqual(computed_result, expected_result)
        computed_result = p.get_theta_names(tex=True)
        expected_result = ["$a$", "$b$"]
        self.assertEqual(computed_result, expected_result)
        # check use case with vector-valued parameter
        p = InferenceProblem("TestProblem")
        p.add_parameter(
            "a",
            "model",
            dim=2,
            tex="$a$",
            prior=("normal", {"loc": [0, 0], "scale": [[1, 0], [0, 1]]}),
        )
        p.add_parameter(
            "b", "model", tex="$b$", prior=("normal", {"loc": 0, "scale": 1})
        )
        computed_result = p.get_theta_names(tex=False, components=False)
        expected_result = ["a", "b"]
        self.assertEqual(computed_result, expected_result)
        computed_result = p.get_theta_names(tex=False, components=True)
        expected_result = ["a_1", "a_2", "b"]
        self.assertEqual(computed_result, expected_result)
        computed_result = p.get_theta_names(tex=True, components=False)
        expected_result = ["$a$", "$b$"]
        self.assertEqual(computed_result, expected_result)
        computed_result = p.get_theta_names(tex=True, components=True)
        expected_result = ["$a_1$", "$a_2$", "$b$"]
        self.assertEqual(computed_result, expected_result)

    def test_theta_explanation(self):
        # simply check that no errors occur when theta_explanation is called; the output
        # that is usually printed is redirected
        p = InferenceProblem("TestProblem")
        p.add_parameter("a", "model", const=1.0)
        p.add_parameter("b", "model", prior=("normal", {"loc": 0, "scale": 1}))
        p.add_parameter("s", "noise", prior=("normal", {"loc": 0, "scale": 1}))
        # check the model_consistency flag
        with self.assertRaises(AssertionError):
            # the model is not consistent
            p.theta_explanation(check_consistency=True)

    def test_add_forward_model(self):
        # check correct use
        p = InferenceProblem("TestProblem")
        p.add_parameter("a", "model", prior=("normal", {"loc": 0, "scale": 1}))
        test_model = ForwardModelBase("a", Sensor("x"), Sensor("y"))
        p.add_forward_model("TestModel", test_model)
        # check for invalid input arguments
        with self.assertRaises(RuntimeError):
            # a given parameter of the forward model has not been added yet
            test_model_2 = ForwardModelBase("b", Sensor("x"), Sensor("y"))
            p.add_forward_model("TestModel_2", test_model_2)
        with self.assertRaises(RuntimeError):
            # add a forward model with the same name
            p.add_forward_model("TestModel", test_model)
        with self.assertRaises(RuntimeError):
            # add a forward model with already used output sensor name
            test_model_2 = ForwardModelBase("a", Sensor("x"), Sensor("y"))
            p.add_forward_model("TestModel_2", test_model_2)

    def test_add_noise_model(self):
        # check correct use
        p = InferenceProblem("TestProblem")
        p.add_parameter("a", "model", prior=("normal", {"loc": 0, "scale": 1}))
        p.add_parameter("s1", "noise", prior=("normal", {"loc": 0, "scale": 1}))
        p.add_parameter("s2", "noise", prior=("normal", {"loc": 0, "scale": 1}))
        p.add_parameter("s3", "noise", prior=("normal", {"loc": 0, "scale": 1}))
        test_model = ForwardModelBase("a", Sensor("x"), [Sensor("y1"), Sensor("y2")])
        p.add_forward_model("TestModel", test_model)
        noise_model1 = NoiseModelBase(
            "normal", "s1", sensors=[Sensor("y1"), Sensor("y2")]
        )
        noise_model2 = NoiseModelBase("normal", ["s2"], sensors=[Sensor("y2")])
        noise_model3 = NoiseModelBase(
            "normal", ["s1", "s2", "s3"], sensors=[Sensor("y1")], name="NM3"
        )
        p.add_noise_model(noise_model1)
        p.add_noise_model(noise_model2)
        p.add_noise_model(noise_model3)
        # adding a noise model with similar sensor interface should not error
        p.add_noise_model(noise_model3)
        # check invalid input arguments
        with self.assertRaises(RuntimeError):
            # the given noise model parameter has not been defined
            p.add_noise_model(
                NoiseModelBase(
                    "normal",
                    "not_existing_parameter",
                    sensors=[Sensor("y1"), Sensor("y2")],
                )
            )

    def test_assign_experiments_to_noise_models(self):
        # some preparations before performing the actual tests
        p = InferenceProblem("TestProblem")
        # add some parameters
        p.add_parameter("a", "model", prior=("normal", {"loc": 0, "scale": 1}))
        p.add_parameter("s1", "noise", prior=("normal", {"loc": 0, "scale": 1}))
        p.add_parameter("s2", "noise", prior=("normal", {"loc": 0, "scale": 1}))
        p.add_parameter("s3", "noise", prior=("normal", {"loc": 0, "scale": 1}))
        # add some dummy forward models (only the output sensors are important)
        test_model_y1 = ForwardModelBase("a", Sensor("x"), Sensor("y1"))
        test_model_y2 = ForwardModelBase("a", Sensor("x"), Sensor("y2"))
        test_model_z1z2 = ForwardModelBase(
            "a", Sensor("x"), [Sensor("z1"), Sensor("z2")]
        )
        p.add_forward_model("TestModel_y1", test_model_y1)
        p.add_forward_model("TestModel_y2", test_model_y2)
        p.add_forward_model("TestModel_z1z2", test_model_z1z2)
        # add some noise models
        noise_model_y1 = NoiseModelBase(
            "normal", ["s1", "s2", "s3"], sensors=[Sensor("y1")]
        )
        noise_model_y2 = NoiseModelBase("normal", ["s2"], sensors=[Sensor("y2")])
        noise_model_y1y2 = NoiseModelBase(
            "normal", "s1", sensors=[Sensor("z1"), Sensor("z2")]
        )
        p.add_noise_model(noise_model_y1)
        p.add_noise_model(noise_model_y2)
        # add some experiments
        p.add_experiment(
            "Exp_y1", sensor_values={"x": 1, "y1": 2}, fwd_model_name="TestModel_y1"
        )
        p.add_experiment(
            "Exp_y2_1", sensor_values={"x": 3, "y2": 4}, fwd_model_name="TestModel_y2"
        )
        p.add_experiment(
            "Exp_y2_2", sensor_values={"x": 5, "y2": 6}, fwd_model_name="TestModel_y2"
        )
        p.add_experiment(
            "Exp_z1z2",
            sensor_values={"x": 7, "z1": 8, "z2": 9},
            fwd_model_name="TestModel_z1z2",
        )
        # so far, not all noise models have been added, so there are unassigned
        # experiments; this should lead to an error
        with self.assertRaises(RuntimeError):
            p.assign_experiments_to_noise_models()
        # now we add the missing noise model
        p.add_noise_model(noise_model_y1y2)
        # this is the call that should be tested here
        p.assign_experiments_to_noise_models()
        # now check if all experiments have been assigned correctly
        self.assertEqual(p.noise_models[0].experiment_names, ["Exp_y1"])
        self.assertEqual(p.noise_models[1].experiment_names, ["Exp_y2_1", "Exp_y2_2"])
        self.assertEqual(p.noise_models[2].experiment_names, ["Exp_z1z2"])

    def test_transform_experimental_data(self):
        # check correct use
        p = InferenceProblem("TestProblem")
        p.add_parameter("a", "model", prior=("normal", {"loc": 0, "scale": 1}))
        forward_model = ForwardModelBase("a", Sensor("x"), Sensor("y"))
        p.add_forward_model("TestModel", forward_model)
        p.add_experiment(
            "Experiment_1",
            fwd_model_name="TestModel",
            sensor_values={"x": [1, 2], "y": [1, 3]},
        )
        # apply a simple power function to the experimental data
        p_copy = p.transform_experimental_data(f=np.power, args=([2, 3],), where=True)
        x_computed = p_copy.experiments["Experiment_1"]["sensor_values"]["x"]
        x_expected = np.array([1, 8])
        self.assertTrue(
            np.allclose(x_computed, x_expected) and x_computed.shape == x_expected.shape
        )
        y_computed = p_copy.experiments["Experiment_1"]["sensor_values"]["y"]
        y_expected = np.array([1, 27])
        self.assertTrue(
            np.allclose(y_computed, y_expected) and y_computed.shape == y_expected.shape
        )


if __name__ == "__main__":
    unittest.main()



# standard library imports
from typing import TYPE_CHECKING
import time
import random
import contextlib

# third party imports
import numpy as np
import emcee
import arviz as az
from loguru import logger
from tabulate import tabulate

# local imports
from probeye.subroutines import pretty_time_delta
from probeye.subroutines import check_for_uninformative_priors
from probeye.inference.scipy_.solver import ScipySolver
from probeye.subroutines import stream_to_logger, print_dict_in_rows

# imports only needed for type hints
if TYPE_CHECKING:  # pragma: no cover
    from probeye.definition.inference_problem import InferenceProblem


class EmceeSolver(ScipySolver):
    """
    Provides emcee-sampler which is a pure-Python implementation of Goodman & Weare’s
    Affine Invariant Markov chain Monte Carlo (MCMC) Ensemble sampler. For more
    information, check out https://emcee.readthedocs.io/en/stable/.
    """

    def __init__(
        self, problem: "InferenceProblem", seed: int = 1, show_progress: bool = True
    ):
        """See docstring of ScipySolver for information on the arguments."""
        logger.debug("Initializing EmceeSolver")
        # check that the problem does not contain a uninformative prior
        check_for_uninformative_priors(problem)
        # initialize the scipy-based solver (ScipySolver)
        super().__init__(problem, seed=seed, show_progress=show_progress)

    def emcee_summary(self, posterior_samples: np.ndarray) -> dict:
        """
        Computes and prints a summary of the posterior samples containing mean, median,
        standard deviation, 5th percentile and 95th percentile. Note, that this method
        was based on code from the taralli package: https://gitlab.com/tno-bim/taralli.

        Parameters
        ----------
        posterior_samples
            The generated samples in an array with as many columns as there are latent
            parameters, and n rows, where n = n_chains * n_steps.

        Returns
        -------
            Keys are the different statistics 'mean', 'median', 'sd' (standard
            deviation), 'q05' and 'q95' (0.05- and 0.95-quantile). The values are
            dictionaries with the parameter names as keys and the respective statistics
            as values.
        """

        # used for the names in the first column
        var_names = self.problem.get_theta_names(tex=False, components=True)

        # compute some stats for each column (i.e., each parameter)
        mean = np.mean(posterior_samples, axis=0)
        quantiles = np.quantile(posterior_samples, [0.50, 0.05, 0.95], axis=0)
        median = quantiles[0, :]
        quantile_05 = quantiles[1, :]
        quantile_95 = quantiles[2, :]

        # compute the sample standard deviations for each parameter
        cov_matrix = np.atleast_2d(np.cov(posterior_samples.T))
        sd = np.sqrt(np.diag(cov_matrix))

        # assemble the summary array
        col_names = ["", "mean", "median", "sd", "5%", "95%"]
        row_names = np.array(var_names)
        tab = np.hstack(
            (
                row_names.reshape(-1, 1),
                mean.reshape(-1, 1),
                median.reshape(-1, 1),
                sd.reshape(-1, 1),
                quantile_05.reshape(-1, 1),
                quantile_95.reshape(-1, 1),
            )
        )

        # print the generated table, and return a summary dict for later use
        print(tabulate(tab, headers=col_names, floatfmt=".2f"))
        return {
            "mean": {name: val for name, val in zip(row_names, mean)},
            "median": {name: val for name, val in zip(row_names, median)},
            "sd": {name: val for name, val in zip(row_names, sd)},
            "q05": {name: val for name, val in zip(row_names, quantile_05)},
            "q95": {name: val for name, val in zip(row_names, quantile_95)},
        }

    def run_mcmc(
        self,
        n_walkers: int = 20,
        n_steps: int = 1000,
        n_initial_steps: int = 100,
        **kwargs,
    ) -> az.data.inference_data.InferenceData:
        """
        Runs the emcee-sampler for the InferenceProblem the EmceeSolver was initialized
        with and returns the results as an arviz InferenceData obj.

        Parameters
        ----------
        n_walkers
            Number of walkers used by the estimator.
        n_steps
            Number of steps to run.
        n_initial_steps
            Number of steps for initial (burn-in) sampling.
        kwargs
            Additional key-word arguments channeled to emcee.EnsembleSampler.

        Returns
        -------
        inference_data
            Contains the results of the sampling procedure.
        """

        # log which solver is used
        logger.info(
            f"Solving problem using emcee sampler with {n_initial_steps} + {n_steps} "
            f"samples and {n_walkers} walkers"
        )
        if kwargs:
            logger.info("Additional options:")
            print_dict_in_rows(kwargs, printer=logger.info)
        else:
            logger.info("No additional options specified")

        # draw initial samples from the parameter's priors
        logger.debug("Drawing initial samples")
        sampling_initial_positions = np.zeros(
            (n_walkers, self.problem.n_latent_prms_dim)
        )
        theta_names = self.problem.get_theta_names(tex=False, components=False)
        for parameter_name in theta_names:
            idx = self.problem.parameters[parameter_name].index
            idx_end = self.problem.parameters[parameter_name].index_end
            samples = self.sample_from_prior(parameter_name, n_walkers)
            if (idx_end - idx) == 1:
                sampling_initial_positions[:, idx] = samples
            else:
                sampling_initial_positions[:, idx:idx_end] = samples

        # The following code is based on taralli and merely adjusted to the variables
        # in the probeye setup; see https://gitlab.com/tno-bim/taralli

        # ............................................................................ #
        #                                 Pre-process                                  #
        # ............................................................................ #

        random.seed(self.seed)
        np.random.seed(self.seed)
        rstate = np.random.mtrand.RandomState(self.seed)

        logger.debug("Setting up EnsembleSampler")
        sampler = emcee.EnsembleSampler(
            nwalkers=n_walkers,
            ndim=self.problem.n_latent_prms_dim,
            log_prob_fn=lambda x: self.logprior(x) + self.loglike(x),
            **kwargs,
        )

        sampler.random_state = rstate

        # ............................................................................ #
        #        Initial sampling, burn-in: used to avoid a poor starting point        #
        # ............................................................................ #

        logger.debug("Starting sampling (initial + main)")
        start = time.time()
        state = sampler.run_mcmc(
            initial_state=sampling_initial_positions,
            nsteps=n_initial_steps,
            progress=self.show_progress,
        )
        sampler.reset()

        # ............................................................................ #
        #                          Sampling of the posterior                           #
        # ............................................................................ #
        sampler.run_mcmc(
            initial_state=state, nsteps=n_steps, progress=self.show_progress
        )
        end = time.time()
        runtime_str = pretty_time_delta(end - start)
        logger.info(
            f"Sampling of the posterior distribution completed: {n_steps} steps and "
            f"{n_walkers} walkers."
        )
        logger.info(f"Total run-time (including initial sampling): {runtime_str}.")
        logger.info("")
        logger.info("Summary of sampling results")
        posterior_samples = sampler.get_chain(flat=True)
        with contextlib.redirect_stdout(stream_to_logger("INFO")):  # type: ignore
            self.summary = self.emcee_summary(posterior_samples)
        self.raw_results = sampler

        # translate the results to a common data structure and return it
        var_names = self.problem.get_theta_names(tex=True, components=True)
        inference_data = az.from_emcee(sampler, var_names=var_names)
        return inference_data



"""
Simple linear regression example with two model and one noise parameter
----------------------------------------------------------------------------------------
The model equation is y = a * x + b with a, b being the model parameters and the noise
model is a normal zero-mean distribution with the std. deviation to infer. The problem
is solved via max likelihood estimation and via sampling using emcee, pyro and dynesty.
"""

# standard library imports
import unittest

# third party imports
import numpy as np
import matplotlib.pyplot as plt

# local imports (problem definition)
from probeye.definition.inference_problem import InferenceProblem
from probeye.definition.forward_model import ForwardModelBase
from probeye.definition.sensor import Sensor
from probeye.definition.noise_model import NormalNoiseModel

# local imports (testing related)
from tests.integration_tests.subroutines import run_inference_engines


class TestProblem(unittest.TestCase):
    def test_linear_regression(
        self,
        n_steps: int = 200,
        n_initial_steps: int = 100,
        n_walkers: int = 20,
        plot: bool = False,
        show_progress: bool = False,
        run_scipy: bool = True,
        run_emcee: bool = True,
        run_torch: bool = True,
        run_dynesty: bool = True,
    ):
        """
        Integration test for the problem described at the top of this file.

        Parameters
        ----------
        n_steps
            Number of steps (samples) to run. Note that the default number is rather low
            just so the test does not take too long.
        n_initial_steps
            Number of steps for initial (burn-in) sampling.
        n_walkers
            Number of walkers used by the estimator.
        plot
            If True, the data and the posterior distributions are plotted. This is
            deactivated by default, so that the test does not stop until the generated
            plots are closed.
        show_progress
            If True, progress-bars will be shown, if available.
        run_scipy
            If True, the problem is solved with scipy (maximum likelihood est).
            Otherwise, no maximum likelihood estimate is derived.
        run_emcee
            If True, the problem is solved with the emcee solver. Otherwise, the emcee
            solver will not be used.
        run_torch
            If True, the problem is solved with the pyro/torch_ solver. Otherwise, the
            pyro/torch_ solver will not be used.
        run_dynesty
            If True, the problem is solved with the dynesty solver. Otherwise, the
            dynesty solver will not be used.
        """

        # ============================================================================ #
        #                              Set numeric values                              #
        # ============================================================================ #

        # 'true' value of a, and its normal prior parameters
        a_true = 2.5
        loc_a = 2.0
        scale_a = 1.0

        # 'true' value of b, and its normal prior parameters
        b_true = 1.7
        loc_b = 1.0
        scale_b = 1.0

        # 'true' value of noise sd, and its uniform prior parameters
        sigma_noise = 0.5
        low_sigma = 0.1
        high_sigma = 0.8

        # the number of generated experiment_names and seed for random numbers
        n_tests = 50
        seed = 1

        # ============================================================================ #
        #                           Define the Forward Model                           #
        # ============================================================================ #

        class LinearModel(ForwardModelBase):
            def response(self, inp: dict) -> dict:
                # this method *must* be provided by the user
                x = inp["x"]
                m = inp["mb"][0]
                b = inp["mb"][1]
                response = {}
                for os in self.output_sensors:
                    response[os.name] = m * x + b
                return response

            def jacobian(self, inp: dict) -> dict:
                # this method *can* be provided by the user; if not provided the
                # jacobian will be approximated by finite differences
                x = inp["x"]  # vector
                one = np.ones(len(x))
                jacobian = {}
                for os in self.output_sensors:
                    # partial derivatives must only be stated for the model parameters;
                    # all other input must be flagged by None; note: partial derivatives
                    # must be given as column vectors
                    jacobian[os.name] = {
                        "x": None,  # x is not a model param.
                        "mb": np.array([x, one]).transpose(),
                    }
                return jacobian

        # ============================================================================ #
        #                         Define the Inference Problem                         #
        # ============================================================================ #

        # initialize the inference problem with a useful name; note that the name will
        # only be stored as an attribute of the InferenceProblem and is not important
        # for the problem itself; can be useful when dealing with multiple problems
        problem = InferenceProblem("Linear regression with normal noise")

        # add all parameters to the problem; the first argument states the parameter's
        # global name (here: 'a', 'b' and 'sigma'); the second argument defines the
        # parameter type (three options: 'model' for parameter's of the forward model,
        # 'prior' for prior parameters and 'noise' for parameters of the noise model);
        # the 'info'-argument is a short description string used for logging, and the
        # tex-argument gives a tex-string of the parameter used for plotting; finally,
        # the prior-argument specifies the parameter's prior; note that this definition
        # of a prior will result in the initialization of constant parameters of type
        # 'prior' in the background
        problem.add_parameter(
            "mb",
            "model",
            dim=2,
            tex="$mb$",
            info="Slope and intercept of the graph",
            prior=(
                "normal",
                {
                    "loc": np.array([loc_a, loc_b]),
                    "scale": np.array([[scale_a, 0], [0, scale_b]]),
                },
            ),
        )
        problem.add_parameter(
            "sigma",
            "noise",
            tex=r"$\sigma$",
            info="Std. dev, of 0-mean noise model",
            prior=("uniform", {"low": low_sigma, "high": high_sigma}),
        )

        # add the forward model to the problem; note that the first positional argument
        # [{'a': 'm'}, 'b'] passed to LinearModel defines the forward model's parameters
        # by name via a list with elements structured like {<global parameter name>:
        # <local parameter name>}; a global name is a name introduced by problem.
        # add_parameter, while a local name is a name used in the response-method of the
        # forward model class (see the class LinearModel above); note that the use of
        # the local parameter name 'm' for the global parameter 'a' is added here only
        # to highlight the possibility of this feature; it is not necessary at all here;
        # whenever forward model's parameter has a similar local and global name (which
        # should be the case most of the times), one doesn't have to use the verbose
        # notation  {<global parameter name>: <local parameter name>} but can instead
        # just write the parameter's (global=local) name, like it is done with the
        # forward model's parameter 'b' below
        isensor = Sensor("x")
        osensor = Sensor("y")
        linear_model = LinearModel(["mb"], [isensor], [osensor])
        problem.add_forward_model("LinearModel", linear_model)

        # add the noise model to the problem
        problem.add_noise_model(
            NormalNoiseModel(prms_def={"sigma": "std"}, sensors=osensor)
        )

        # ============================================================================ #
        #                    Add test data to the Inference Problem                    #
        # ============================================================================ #

        # data-generation; normal noise with constant variance around each point
        np.random.seed(seed)
        x_test = np.linspace(0.0, 1.0, n_tests)
        y_true = linear_model.response({isensor.name: x_test, "mb": [a_true, b_true]})[
            osensor.name
        ]
        y_test = np.random.normal(loc=y_true, scale=sigma_noise)

        # add the experimental data
        problem.add_experiment(
            f"TestSeries_1",
            fwd_model_name="LinearModel",
            sensor_values={isensor.name: x_test, osensor.name: y_test},
        )

        # give problem overview
        problem.info()

        # plot the true and noisy data
        if plot:
            plt.scatter(x_test, y_test, label="measured data", s=10, c="red", zorder=10)
            plt.plot(x_test, y_true, label="true", c="black")
            plt.xlabel(isensor.name)
            plt.ylabel(osensor.name)
            plt.legend()
            plt.tight_layout()
            plt.draw()  # does not stop execution

        # ============================================================================ #
        #                    Solve problem with inference engine(s)                    #
        # ============================================================================ #

        # this routine is imported from another script because it it used by all
        # integration tests in the same way; ref_values are used for plotting
        true_values = {"mb": [a_true, b_true], "sigma": sigma_noise}
        run_inference_engines(
            problem,
            true_values=true_values,
            n_steps=n_steps,
            n_initial_steps=n_initial_steps,
            n_walkers=n_walkers,
            plot=plot,
            show_progress=show_progress,
            run_scipy=run_scipy,
            run_emcee=run_emcee,
            run_torch=run_torch,
            run_dynesty=run_dynesty,
        )


if __name__ == "__main__":
    unittest.main()



# standard library imports
import unittest

# third party imports
import numpy as np

# local imports
from probeye.definition.forward_model import ForwardModelBase
from probeye.definition.sensor import Sensor


class TestProblem(unittest.TestCase):
    def test_undefined_forward_model(self):
        # check for the not-implemented-error when no response method is defined
        forward_model = ForwardModelBase(["a"], Sensor("x"), Sensor("y"))
        with self.assertRaises(NotImplementedError):
            forward_model.response({})
        with self.assertRaises(NotImplementedError):
            forward_model({})

    def test_model_template_one_sensor(self):

        # define a simple model using ForwardModelBase
        class ForwardModel(ForwardModelBase):
            def response(self, inp):
                x = inp["x"]
                a = inp["a"]
                b = inp["b"]
                return {"y": a * x ** 2 + b}

        # check the __call__-method
        forward_model = ForwardModel(["a", "b"], [Sensor("x")], [Sensor("y")])
        prms = {"a": 1, "b": 2}
        computed_result = forward_model({**{"x": 1.0}, **prms})
        expected_result = {"y": 3.0}
        self.assertEqual(computed_result, expected_result)

        # check input/output sensor names
        self.assertEqual(forward_model.input_sensor_names, ["x"])
        self.assertEqual(forward_model.output_sensor_names, ["y"])

        # check the jacobian-method (dict-version)
        computed_result = forward_model.jacobian({**{"x": 1.0}, **prms})
        expected_result = {
            "y": {"x": None, "a": np.array([[1.0]]), "b": np.array([[1.0]])}
        }
        for k1, v1 in computed_result.items():
            for k2, v2 in v1.items():
                self.assertAlmostEqual(v2, expected_result[k1][k2], places=2)
        # check the jacobian-method (array-version)
        inp_ = {**{"x": 1.0}, **prms}
        jac_dict = forward_model.jacobian(inp_)
        computed_result = forward_model.jacobian_dict_to_array(inp_, jac_dict, 3)
        # note that the first element of the expected result is zero, because it wasn't
        # computed; the first input channel is not a model parameter, hence its partial
        # derivative is not evaluated
        expected_result = np.array([[0.0, 1.0, 1.0]])
        self.assertTrue(
            np.allclose(computed_result, expected_result, atol=1e-3)
            and computed_result.shape == expected_result.shape
        )

    def test_model_template_multiple_sensors(self):

        # define an output sensor with an offset attribute
        class OutputSensorOffset(Sensor):
            def __init__(self, name, offset):
                super().__init__(name)
                self.offset = offset

        # define a simple model using ForwardModelBase
        class ForwardModel(ForwardModelBase):
            def response(self, inp):
                x1 = inp["x1"]
                x2 = inp["x2"]
                a = inp["a"]
                b = inp["b"]
                response_dict = dict()
                for os in self.output_sensors:
                    response_dict[os.name] = a * x1 ** 2 + b * x2 + os.offset
                return response_dict

        # check the __call__-method
        is1, is2 = Sensor("x1"), Sensor("x2")
        os1, os2 = OutputSensorOffset("y1", 0.1), OutputSensorOffset("y2", -0.2)
        forward_model = ForwardModel(["a", "b"], [is1, is2], [os1, os2])
        prms = {"a": 1, "b": 2}
        computed_result = forward_model({**{"x1": 2.0, "x2": 3.0}, **prms})
        expected_result = {"y1": 10.1, "y2": 9.8}
        self.assertEqual(computed_result, expected_result)

        # check input/output sensor names
        self.assertEqual(forward_model.input_sensor_names, ["x1", "x2"])
        self.assertEqual(forward_model.output_sensor_names, ["y1", "y2"])

        # check the jacobian-method (dict-version)
        computed_result = forward_model.jacobian({**{"x1": 2.0, "x2": 3.0}, **prms})
        expected_result = {
            "y1": {
                "x1": None,
                "x2": None,
                "a": np.array([[4.0]]),
                "b": np.array([[3.0]]),
            },
            "y2": {
                "x1": None,
                "x2": None,
                "a": np.array([[4.0]]),
                "b": np.array([[3.0]]),
            },
        }
        for k1, v1 in computed_result.items():
            for k2, v2 in v1.items():
                self.assertAlmostEqual(v2, expected_result[k1][k2], places=2)
        # check the jacobian-method (array-version)
        inp_ = {**{"x1": 2.0, "x2": 3.0}, **prms}
        jac_dict = forward_model.jacobian(inp_)
        computed_result = forward_model.jacobian_dict_to_array(inp_, jac_dict, 4)
        # note that the first two elements of the expected result-rows are zero, because
        # they weren't computed; the first two input channels are not model parameters,
        # hence their partial derivatives are not evaluated
        expected_result = np.array([[0.0, 0.0, 4.0, 3.0], [0.0, 0.0, 4.0, 3.0]])
        self.assertTrue(
            np.allclose(computed_result, expected_result, atol=1e-3)
            and computed_result.shape == expected_result.shape
        )


if __name__ == "__main__":
    unittest.main()



# standard library imports
from typing import Union, List, Tuple, Optional, TYPE_CHECKING
import copy as cp

# third party imports
import copy
import numpy as np
import scipy as sp
from scipy.optimize import minimize
from loguru import logger

# local imports
from probeye.inference.scipy_.priors import translate_prior
from probeye.inference.scipy_.noise_models import translate_noise_model
from probeye.subroutines import print_dict_in_rows, make_list

# imports only needed for type hints
if TYPE_CHECKING:  # pragma: no cover
    from probeye.definition.inference_problem import InferenceProblem


class ScipySolver:
    """Solver routines based on scipy and numpy for an InferenceProblem."""

    def __init__(
        self, problem: "InferenceProblem", seed: int = 1, show_progress: bool = True
    ):
        """
        Parameters
        ----------
        problem
            Describes the inference problem including e.g. parameters and data.
        seed
            Random state used for random number generation.
        show_progress
            When True, the progress of a solver routine will be shown (for example as a
            progress-bar) if such a feature is available. Otherwise, the progress will
            not shown.
        """

        # log at beginning so that errors can be associated
        logger.debug("Initializing ScipySolver")

        # attributes from arguments
        self.show_progress = show_progress
        self.seed = seed

        # the following attributes will be set after the solver was run
        self.raw_results = None
        self.summary = {}  # type: dict

        # each noise model must be connected to the relevant experiment_names; a copy is
        # created before, so that this solver routine does not have side effects on the
        # original problem; such side effects would occur due to calling the assign_
        # experiments_to_noise_models-method below
        try:
            self.problem = cp.deepcopy(problem)
        except:
            logger.warning(
                "The inference problem could not be deep-copied! The original problem "
                "will be modified!"
            )
            self.problem = problem
        self.problem.assign_experiments_to_noise_models()

        # translate the prior definitions to objects with computing capabilities
        logger.debug("Translate problem's priors")
        self.priors = copy.deepcopy(self.problem.priors)
        for prior_name, prior_template in self.problem.priors.items():
            self.priors[prior_name] = translate_prior(prior_template)

        # translate the general noise model objects into solver specific ones
        logger.debug("Translate problem's noise models")
        self.noise_models = []
        for noise_model_base in self.problem.noise_models:
            self.noise_models.append(translate_noise_model(noise_model_base))

    def evaluate_model_response(
        self, theta: np.ndarray, experiment_names: Union[str, List[str], None] = None
    ) -> dict:
        """
        Evaluates the model response for each forward model for the given parameter
        vector theta and the given experiments.

        Parameters
        ----------
        theta
            A numeric vector for which the model responses should be evaluated. Which
            parameters these numbers refer to can be checked by calling self.theta_
            explanation() once the problem is set up.
        experiment_names
            Contains the names of all or some of the experiments added to the inference
            problem. If this argument is None (which is a common use case) then all
            experiments defined in the problem (self.experiments) are used. The names
            provided here define the experiments that the fwd. model is evaluated for.

        Returns
        -------
        model_response_dict
            The first key is the name of the experiment. The values are dicts which
            contain the forward model's output sensor's names as keys have the
            corresponding model responses as values.
        """

        # if experiments is not further specified all experiments added to the problem
        # will be accounted for when computing the model error
        if experiment_names is None:
            experiment_names = [*self.problem.experiments.keys()]
        else:
            # make sure that a given string is converted into a list
            experiment_names = make_list(experiment_names)

        # first, loop over all forward models, and then, over all experiments that are
        # associated with the corresponding model
        model_response_dict = {}
        for fwd_name, forward_model in self.problem.forward_models.items():
            # get the model parameters for the considered forward model
            prms_model = self.problem.get_parameters(theta, forward_model.prms_def)
            # get all experiments referring to the considered forward model
            relevant_experiment_names = self.problem.get_experiment_names(
                forward_model_names=fwd_name, experiment_names=experiment_names
            )
            # evaluate the forward model for each relevant experiment
            for exp_name in relevant_experiment_names:
                exp_dict = self.problem.experiments[exp_name]
                # prepare the model input values from the experimental data
                sensor_values = exp_dict["sensor_values"]
                exp_inp = {
                    input_sensor.name: sensor_values[input_sensor.name]
                    for input_sensor in forward_model.input_sensors
                }
                inp = {**exp_inp, **prms_model}  # adds the two dictionaries
                # finally, evaluate the forward model for this experiment
                model_response_dict[exp_name] = forward_model(inp)

        return model_response_dict

    def logprior(self, theta: np.ndarray) -> float:
        """
        Evaluates the log-prior function of the problem at theta.

        Parameters
        ----------
        theta
            A numeric vector for which the log-likelihood function should be evaluated.
            Which parameters these numbers refer to can be checked by calling self.
            theta_explanation() once the problem is set up.

        Returns
        -------
        lp
            The evaluated log-prior function for the given theta-vector.
        """
        lp = 0.0
        for prior in self.priors.values():
            prms = self.problem.get_parameters(theta, prior.prms_def)
            lp += prior(prms, "logpdf")
        return lp

    def sample_from_prior(self, prm_name: str, size: int) -> np.ndarray:
        """
        Generates random samples from a parameter's prior distribution and returns the
        generated samples.

        Parameters
        ----------
        prm_name
            The name of the parameter the prior is associated with.
        size
            The number of random samples to be drawn.

        Returns
        -------
            The generated samples.
        """
        prior = self.priors[self.problem.parameters[prm_name].prior.name]
        # check for prior-priors; if a prior parameter is a latent parameter and not a
        # constant, one first samples from the prior parameter's prior distribution, and
        # then takes the mean of those samples to sample from the first prior
        # distribution; this procedure is recursive, so that (in principle) one could
        # also define priors of the prior's prior parameters and so forth
        theta_aux = np.zeros(self.problem.parameters.n_latent_prms)
        for prior_prm_name in prior.hyperparameters.keys():
            if self.problem.parameters[prior_prm_name].role == "latent":
                samples = self.sample_from_prior(prior_prm_name, size)
                theta_aux[self.problem.parameters[prior_prm_name].index] = np.mean(
                    samples
                )
        prms = self.problem.get_parameters(theta_aux, prior.hyperparameters)
        return prior.generate_samples(prms, size)

    def loglike(self, theta: np.ndarray) -> float:
        """
        Evaluates the log-likelihood function of the problem at theta.

        Parameters
        ----------
        theta
            A numeric vector for which the log-likelihood function should be evaluated.
            Which parameters these numbers refer to can be checked by calling self.
            theta_explanation() once the problem is set up.

        Returns
        -------
        ll
            The evaluated log-likelihood function for the given theta-vector.
        """
        # compute the contribution to the log-likelihood function for each noise
        # model and sum it all up
        ll = 0.0
        for noise_model in self.noise_models:
            # compute the model response for the noise model's experiment_names
            model_response = self.evaluate_model_response(
                theta, noise_model.experiment_names
            )
            # get the parameter values for the noise model's parameters
            prms_noise = self.problem.get_parameters(theta, noise_model.prms_def)
            # evaluate the loglike-contribution for the noise model
            ll += noise_model.loglike_contribution(model_response, prms_noise)
        return ll

    def get_start_values(
        self,
        x0_dict: Optional[dict] = None,
        x0_prior: str = "mean",
        x0_default: float = 1.0,
    ) -> Tuple[np.ndarray, dict]:
        """
        Derives the start values for the maximum likelihood optimization run. For an
        explanation of the arguments, see self.run_max_likelihood.

        Returns
        -------
        x0
            A numeric vector with the derived start values in the order of
            InferenceProblem.get_theta_names().
        x0_dict
            Keys are the latent parameters, while the keys are their start values.
        """

        # this is going to be the start vector
        x0 = np.zeros(self.problem.n_latent_prms_dim)

        if x0_dict:
            # in this case, the user explicitly defined the start values
            for prm_name, prm_value in x0_dict.items():
                idx = self.problem.parameters[prm_name].index
                idx_end = self.problem.parameters[prm_name].index_end
                x0[idx:idx_end] = prm_value
        else:
            # in this case, the start values are derived from the priors; if a prior is
            # not uninformative, its mean value will be used; if a prior is
            # uninformative, the x0_default value will be used
            x0_dict = {}
            prms = cp.copy(self.problem.constant_prms_dict)
            for prm_name in self.problem.get_theta_names():
                prior_name = self.problem.parameters[prm_name].prior.name
                prior_type = self.problem.parameters[prm_name].prior.prior_type
                idx = self.problem.parameters[prm_name].index
                idx_end = self.problem.parameters[prm_name].index_end
                dim = self.problem.parameters[prm_name].dim
                if prior_type != "uninformative":
                    prm_value = self.priors[prior_name](
                        prms, x0_prior, use_ref_prm=False
                    )
                    prms[prm_name] = prm_value
                    x0[idx:idx_end] = prm_value
                else:
                    # no mean value can be requested if the prior is uninformative,
                    # hence a default value is used
                    x0[idx:idx_end] = [x0_default] * dim
                # scalar values should not be saved as one-element-lists
                if dim == 1:
                    x0_dict[prm_name] = x0[idx]
                else:
                    x0_dict[prm_name] = x0[idx:idx_end]

        return x0, x0_dict

    def summarize_ml_results(
        self,
        minimize_results: sp.optimize.optimize.OptimizeResult,
        true_values: Optional[dict],
        x0_dict: dict,
    ):
        """
        Prints a summary of the results of the maximum likelihood estimation. For an
        explanation of the arguments, check out the docstring of the self.run_max_
        likelihood-method.
        """

        # the first part of the summary contains process information
        n_char_message = len(minimize_results.message)
        msg = (
            f"\nResults of maximum likelihood estimation\n"
            f"{'=' * n_char_message}\n"
            f"{minimize_results.message}\n"
            f"{'-' * n_char_message}\n"
            f"Number of iterations:           {minimize_results.nit}\n"
            f"Number of function evaluations: {minimize_results.nfev}\n"
            f"{'-' * n_char_message}"
        )
        for line in msg.split("\n"):
            logger.info(line)

        # the second part shows the actual results and compares them with the true
        # values (if given) and the start values
        if minimize_results.success:
            theta_names = self.problem.get_theta_names(tex=False)
            n_char = max([len(name) for name in theta_names]) + 4
            for theta_name in theta_names:
                idx = self.problem.parameters[theta_name].index
                idx_end = self.problem.parameters[theta_name].index_end
                opt_name = f"{theta_name}_opt"
                opt_val = minimize_results.x[idx:idx_end]
                line = f"{opt_name:{n_char}s} = {opt_val}"
                if true_values:
                    line += (
                        f" (true = {true_values[theta_name]}, "
                        f"start = {x0_dict[theta_name]})"
                    )
                else:
                    line += f" (start = {x0_dict[theta_name]})"
                logger.info(line)
        logger.info("")  # empty line for visual buffer

    def run_max_likelihood(
        self,
        x0_dict: Optional[dict] = None,
        x0_prior: str = "mean",
        x0_default: float = 1.0,
        true_values: Optional[dict] = None,
        method: str = "Nelder-Mead",
        solver_options: Optional[dict] = None,
    ) -> sp.optimize.optimize.OptimizeResult:
        """
        Finds values for an InferenceProblem's latent parameters that maximize the
        problem's likelihood function. The used method is scipy's minimize function from
        the optimize submodule.

        Parameters
        ----------
        x0_dict
            Contains the start values for each latent variable. Via this arg the user
            can explicitly specify a start value for the optimization.
        x0_prior
            If x0_dict is not given, the start values will be derived from the priors,
            either using the 'mean' or 'median' value. If x0_dict is given, this
            argument has no effect. Valid values are 'mean' and 'median'.
        x0_default
            For uninformative priors, no mean or median value is defined. In those
            cases, the default_x0 value will be used as start value. If x0_dict is
            given, this argument has no effect.
        true_values
            Defines 'true' parameter values. Keys are the parameter names and values are
            the 'true' values. They are only used to print them next to the inferred
            parameter values from the optimization run.
        method
            Defines the algorithm used by scipy.optimize.minimize. See the documentation
            of this scipy method to see all the options.
        solver_options
            Options passed to scipy.optimize.minimize under the 'options' keyword arg.
            See the documentation of this scipy method to see available options.

        Returns
        -------
        minimize_results
            An object returns by scipy's minimize function containing the optimization
            results. The parameter vector that optimizes the likelihood function can be
            requested via 'minimize_results.x'.
        """

        # log at beginning so that errors can be associated
        logger.info("Solving problem via maximum likelihood estimation")

        # since scipy's minimize function is used, we need a function that returns the
        # negative log-likelihood function (minimizing the negative log-likelihood is
        # equivalent to maximizing the (log-)likelihood)
        def fun(x):
            return -self.loglike(x)

        # prepare the start value either from the given x0_dict or from the mean values
        # of the latent parameter's priors
        logger.debug("Deriving start values")
        x0, x0_dict = self.get_start_values(
            x0_dict=x0_dict, x0_prior=x0_prior, x0_default=x0_default
        )
        logger.info("Using start values:")
        print_dict_in_rows(x0_dict, printer=logger.info, val_fmt=None)

        # this is the where the solver does its thing
        logger.info(f"Starting optimizer (using {method})")
        if solver_options:
            logger.info("Specified solver options:")
            print_dict_in_rows(solver_options, printer=logger.info)
        else:
            logger.info("No solver options specified")
        minimize_results = minimize(fun, x0, method=method, options=solver_options)

        # note that in this case, the raw solver result is identical with the return-
        # value of this method; however, for other solver they differ; hence, this
        # attribute is set here only for consistency reasons
        self.raw_results = minimize_results
        self.summary = {
            "success": minimize_results.success,
            "theta_opt": minimize_results.x,
        }

        # some convenient printout with respect to the solver's results
        self.summarize_ml_results(minimize_results, true_values, x0_dict)

        return minimize_results



# standard library imports
import unittest

# local imports
from probeye.definition.prior import PriorBase


class TestProblem(unittest.TestCase):
    def test_prior_template(self):
        prior_template = PriorBase(
            "a", ["loc_a", "scale_a"], "a_normal", "normal distribution"
        )
        # check that the attributes are wired correctly
        self.assertEqual(prior_template.ref_prm, "a")
        self.assertEqual(prior_template.name, "a_normal")
        self.assertEqual(prior_template.prior_type, "normal distribution")
        self.assertEqual(
            prior_template.prms_def, {"a": "a", "loc_a": "loc_a", "scale_a": "scale_a"}
        )
        self.assertEqual(
            prior_template.hyperparameters, {"loc_a": "loc_a", "scale_a": "scale_a"}
        )
        # check that the __str__-method works
        self.assertTrue(len(prior_template.__str__()) > 0)


if __name__ == "__main__":
    unittest.main()



# standard library imports
import unittest

# third party imports
from scipy import stats
import numpy as np

# local imports
from probeye.definition.sensor import Sensor
from probeye.inference.scipy_.noise_models import NormalNoise


class TestProblem(unittest.TestCase):
    def test_normal_noise_init(self):
        x_test = np.linspace(0.0, 1.0, 10)
        osensor = Sensor("y", x=x_test)
        with self.assertRaises(ValueError):
            NormalNoise(
                osensor,
                prms_def=[{"s": "std"}, "l_corr"],
                corr="x",
                corr_model="wrong model spec",
                noise_type="additive",
            )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import setuptools
import site
import sys

# https://github.com/googlefonts/fontmake/commit/164b24fd57c062297bf68b21c8ae88bc676f090b
site.ENABLE_USER_SITE = "--user" in sys.argv[1:]

if __name__ == "__main__":
    setuptools.setup()



# standard library
from typing import Optional, TYPE_CHECKING

# third party imports
import matplotlib.pyplot as plt

# local imports (inference engines)
from probeye.inference.scipy_.solver import ScipySolver
from probeye.inference.emcee_.solver import EmceeSolver
from probeye.inference.torch_.solver import PyroSolver
from probeye.inference.dynesty_.solver import DynestySolver

# local imports (post-processing)
from probeye.postprocessing.sampling import create_pair_plot
from probeye.postprocessing.sampling import create_posterior_plot
from probeye.postprocessing.sampling import create_trace_plot

# imports only needed for type hints
if TYPE_CHECKING:
    from probeye.definition.inference_problem import InferenceProblem


def run_inference_engines(
    problem: "InferenceProblem",
    true_values: Optional[dict] = None,
    n_steps: int = 1000,
    n_initial_steps: int = 100,
    n_walkers: int = 20,
    plot: bool = True,
    show_progress: bool = True,
    run_scipy: bool = True,
    run_emcee: bool = True,
    run_torch: bool = True,
    run_dynesty: bool = True,
):
    """
    Runs a requested selection of inference engines on a given problem. This function is
    in a separate file in order to avoid repeating the same lines of code in all the
    integration tests.

    Parameters
    ----------
    problem
        Describes the inference problem including e.g. parameters and data.
    true_values
        Used for plotting 'true' parameter values. Keys are the parameter names and
        values are the values that are supposed to be shown in the marginal plots.
    n_steps
        Number of steps (samples) to run.
    n_initial_steps
        Number of steps for initial (burn-in) sampling.
    n_walkers
        Number of walkers used by the estimator.
    plot
        If True, the data and the post-processing plots are plotted.
    show_progress
        If True, progress-bars will be shown, if available.
    run_scipy
        If True, the problem is solved with scipy (maximum likelihood estimate).
        Otherwise, no maximum likelihood estimate is derived.
    run_emcee
        If True, the problem is solved with the emcee solver. Otherwise, the
        emcee solver will not be used.
    run_torch
        If True, the problem is solved with the pyro/torch_ solver. Otherwise, the
        pyro/torch_ solver will not be used.
    run_dynesty
        If True, the problem is solved with the dynesty solver. Otherwise, the
        dynesty solver will not be used.
    """

    def create_plots(inference_data, problem, true_values):
        create_pair_plot(
            inference_data,
            problem,
            true_values=true_values,
            show=False,
            title="plot_priors=True, focus_on_posterior=False (default)",
        )
        create_pair_plot(
            inference_data,
            problem,
            focus_on_posterior=True,
            kind="hexbin",
            true_values=true_values,
            show=False,
            marginal_kwargs={"kind": "hist", "hist_kwargs": {"bins": 10}},
            title="plot_priors=True, focus_on_posterior=True (hex + hist)",
        )
        create_pair_plot(
            inference_data,
            problem,
            plot_priors=False,
            kind="scatter",
            true_values=true_values,
            show=False,
            marginal_kwargs={"kind": "hist", "hist_kwargs": {"bins": 10}},
            title="plot_priors=False (scatter + hist)",
        )
        create_pair_plot(
            inference_data,
            problem,
            plot_priors=False,
            true_values=true_values,
            show=False,
            title="plot_priors=False",
        )
        create_posterior_plot(
            inference_data,
            problem,
            true_values=true_values,
            show=False,
            title="This is a posterior-plot",
        )
        create_trace_plot(
            inference_data, problem, show=False, title="This is a trace-plot"
        )
        if plot:
            plt.show()  # shows all plots at once due to 'show=False' above

    if run_scipy:
        scipy_solver = ScipySolver(problem, show_progress=show_progress)
        inference_data = scipy_solver.run_max_likelihood(true_values=true_values)
        # no post processing for scipy solver as that is no sampler

    if run_emcee:
        emcee_solver = EmceeSolver(problem, show_progress=show_progress)
        inference_data = emcee_solver.run_mcmc(
            n_walkers=n_walkers,
            n_steps=n_steps,
            n_initial_steps=n_initial_steps,
        )
        create_plots(inference_data, problem, true_values)

    if run_torch:
        n_walkers_used = 1  # getting errors when trying to use more
        pyro_solver = PyroSolver(problem, show_progress=show_progress)
        inference_data = pyro_solver.run_mcmc(
            n_walkers=n_walkers_used,
            n_steps=n_steps,
            n_initial_steps=n_initial_steps,
        )
        create_plots(inference_data, problem, true_values)

    if run_dynesty:
        dynesty_solver = DynestySolver(problem, show_progress=show_progress)
        inference_data = dynesty_solver.run_dynesty("static", nlive=250)
        create_plots(inference_data, problem, true_values)



# standard library imports
from copy import copy
from typing import Union, List, Tuple, Any, Optional, Generator, Callable
from typing import TYPE_CHECKING
import os
import sys

# third party imports
import numpy as np
from loguru import logger

# local imports for type checking
if TYPE_CHECKING:  # pragma: no cover
    from probeye.definition.inference_problem import InferenceProblem


def len_or_one(obj: Any) -> int:
    """
    Returns the length of an object or 1 if no length is defined.

    Parameters
    ----------
    obj
        Most of the time this will be a list/tuple or a single scalar number. But
        generally, it can be anything.

    Returns
    -------
        The length of the given list/tuple etc. or 1, if obj has no __len__-attribute;
        the latter case is mostly intended for scalar numbers.

    """
    if hasattr(obj, "__len__"):
        # the following check is necessary, since the len-function applied to a numpy
        # array of format numpy.array(1) results in a TypeError
        if type(obj) is np.ndarray:
            if not obj.shape:
                return 1
            else:
                return len(obj)
        else:
            return len(obj)
    else:
        return 1


def make_list(arg: Any) -> list:
    """
    Converts a given argument into a list, if it is not a list or tuple. The typical use
    case for this method is to convert a single string into a list with this string as
    its only element, e.g. make_list('sigma') = ['sigma'].

    Parameters
    ----------
    arg
        Essentially anything. Most of the time this might be a string or a list of
        strings or a tuple of strings.

    Returns
    -------
    new_arg
        Either arg if it is of type list or tuple, or a list with arg as its only
        element if arg is not of type list or tuple.
    """
    if type(arg) in [list, tuple]:
        new_arg = arg
    else:
        new_arg = [copy(arg)]
    return new_arg


def underlined_string(
    string: str, symbol: str = "=", n_empty_start: int = 1, n_empty_end: int = 1
) -> str:
    """
    Adds a line made of 'symbol'-characters under a given string and returns it.

    Parameters
    ----------
    string
        The string that should be underlined.
    symbol
        A single character the line should be 'made' of.
    n_empty_start
        Number of empty lines added before the underlined string.
    n_empty_end
        Number of empty lines added after the underlined string.

    Returns
    -------
    result_string
        The generated string representing an underlined string, possibly with empty
        lines added before/after.
    """
    n_chars = len(string)
    underline_string = n_chars * symbol
    empty_lines_start = n_empty_start * "\n"
    empty_lines_end = n_empty_end * "\n"
    result_string = string + "\n" + underline_string
    result_string = empty_lines_start + result_string + empty_lines_end
    return result_string


def titled_table(
    title_str: str,
    table_str: str,
    symbol: str = "-",
    n_empty_start: int = 1,
    n_empty_end: int = 0,
) -> str:
    """
    Adds an underlined title string to a given table string. The line, that underlines
    the title will be as long as the longest line of the table.

    Parameters
    ----------
    title_str
        The title to be put on top of the table.
    table_str
        The string representing the table. For example generated using tabulate.
    symbol
        A single character the line should be 'made' of.
    n_empty_start
        Number of empty lines added before the title.
    n_empty_end
        Number of empty lines added after the table string.

    Returns
    -------
    result_string
        An underlined title, followed by a table.
    """
    # get the number of characters in the given table's longest line
    max_line_length = max([len(line) for line in table_str.split("\n")])
    # now, simply concatenate the different lines
    result_string = (
        n_empty_start * "\n"
        + title_str
        + "\n"
        + max_line_length * symbol
        + "\n"
        + table_str
        + "\n"
        + n_empty_end * "\n"
    )
    return result_string


def replace_string_chars(
    string: str, replace: Optional[dict] = None, remove: Optional[list] = None
) -> str:
    """
    Removes and replaces characters from a given string according to the input.

    Parameters
    ----------
    string
        The string to be modified.
    replace
        The keys are the characters to be replaced, the values are stating their
        replacements.
    remove
        A list of characters to be removed from the given string.

    Returns
    -------
     string
        The modified string with removed/replaced characters.
    """
    # first, do the replacements
    if replace is not None:
        for char, replacement in replace.items():
            string = string.replace(char, replacement)
    # finally, remove characters as requested
    if remove is not None:
        for char in remove:
            string = string.replace(char, "")
    return string


def simplified_list_string(list_: list) -> str:
    """
    Given a list, it returns the string returned by its __str__ method, where some
    characters are removed for a slightly easier to read printout. Example: the list
    ['a', 1.2] is usually printed as '['a', 1.2]'. Here, it will be converted into the
    string 'a, 1.2' (no brackets, and no quotes).

    Parameters
    ----------
    list_
        Some list to be printed.

    Returns
    -------
    simplified_list_str
        The list_'s __str__ method's return string without brackets and quotes.
    """
    simplified_list_str = replace_string_chars(str(list_), remove=["[", "]", "'"])
    return simplified_list_str


def simplified_dict_string(dict_: dict) -> str:
    """
    Given a dictionary, it returns the string returned by its __str__ method, where some
    characters are removed for a slightly easier to read printout. For example: the dict
    {'a': 1.2} is usually printed as '{'a': 1.2, 'b': 2.1}'. Here, it will be converted
    into the string 'a=1.2, b=2.1'.

    Parameters
    ----------
    dict_
        Some dictionary to be printed.

    Returns
    -------
    simplified_dict_str
        Modified version of dict_'s __str__ method's return string (no quotes, no braces
        and the colon will be replaced with an equal sign.
    """
    simplified_dict_str = replace_string_chars(
        str(dict_), remove=["{", "}", "'"], replace={": ": "="}
    )
    return simplified_dict_str


def unvectorize_dict_values(dict_: dict) -> list:
    """
    Takes a dict with items like <name>: <vector> and converts it into a list, where
    each element is a 'fraction' or the whole dictionary. The following example will
    illustrate it: {'x': [1, 2, 3], 'y': [4, 5, 6]} will be converted into
    [{'x': 1, 'y': 4}, {'x': 2, 'y': 5}, {'x': 3, 'y': 6}].

    Parameters
    ----------
    dict_
        The dictionary that should be converted. All values must be 1D arrays of the
        same length.

    Returns
    -------
    result_list
        The 'un-vectorized' dictionary. Check out the example above.
    """

    # all values must be iterable
    dict_copy = copy(dict_)
    for key, value in dict_.items():
        if not hasattr(value, "__len__"):
            dict_copy[key] = [value]

    # check if all lengths are the same
    if len({len(vector) for vector in dict_copy.values()}) != 1:
        raise RuntimeError("The values of the dictionary have different lengths!")

    # create the result list
    vector_length = len([*dict_copy.values()][0])
    keys = [*dict_.keys()]
    result_list = []
    for i in range(vector_length):
        atom_dict = dict()
        for key in keys:
            atom_dict[key] = dict_copy[key][i]
        result_list.append(atom_dict)

    return result_list


def sub_when_empty(string: str, empty_str: str = "-") -> str:
    """
    Just returns a given string if it is not empty. If it is empty though, a default
    string is returned instead.

    Parameters
    ----------
    string
        The string to check if it is empty or not
    empty_str
        The string to be returned if the given string is empty

    Returns
    -------
    result_string
        Either the given string (when 'string' is not empty) or the empty_str (when
        'string' is empty)
    """
    if type(string) is not str:
        raise TypeError(f"Input must be of type string. Found: type '{type(string)}'")
    if len(string) > 0:
        result_string = string
    else:
        result_string = empty_str
    return result_string


def dict2list(dict_: dict) -> list:
    """
    Converts a dict into a list of key-value dictionaries and returns it.

    Parameters
    ----------
    dict_
        Some dictionary to be converted.

    Returns
    -------
    list_
        Each element is a dict with one key-value pair. These key-value pairs
        are those contained in dict_.
    """
    if type(dict_) != dict:
        raise TypeError(
            f"Input argument must be of type 'dict', found '{type(dict_)}'."
        )
    list_ = []
    for key, value in dict_.items():
        list_.append({key: value})
    return list_


def list2dict(list_dict: Union[list, dict]) -> dict:
    """
    Converts a list into a specific dictionary. The list may only contain strings or
    one-element dictionaries. For example [{'a': 'm'}, 'b'] will be converted into
    {'a': 'm', 'b': 'b'}.

    Parameters
    ----------
    list_dict
        If it's a list it may only contain strings or one-element dictionaries.

    Returns
    -------
    dict_
        Strings are mapped to themselves, while one-element dictionaries are simply
        added to this result dictionary.
    """
    # check the given input
    if type(list_dict) not in [list, dict]:
        raise TypeError(
            f"The input argument must be of type 'list' or 'dict'. Found type "
            f"'{type(list_dict)}' however."
        )
    if type(list_dict) is dict:
        # convert the dict to a list, so it can be checked by this function
        list_ = dict2list(copy(list_dict))  # type: Union[list, dict]
    else:
        list_ = copy(list_dict)
    # convert the list to a dictionary
    dict_ = {}  # type: dict
    for element in list_:
        element_type = type(element)
        if element_type == dict:
            if len(element) != 1:
                raise ValueError(
                    f"Found a dict-element, which has {len(element)} instead "
                    f"of 1 key-value pair."
                )
            dict_ = {**dict_, **element}
        elif element_type == str:
            dict_[element] = element
        else:
            raise TypeError(
                f"The elements in the given list must be either of type "
                f"'string' or 'dict'. Found '{element_type}' however."
            )
    return dict_


def pretty_time_delta(seconds: Union[float, int]) -> str:
    """
    Converts number of seconds into a human friendly time string. Source: https:
    //gist.github.com/thatalextaylor/7408395#file-1-python-pretty-time-delta-py

    Parameters
    ----------
    seconds
        The given number of seconds to be converted,.

    Returns
    -------
        Human friendly time difference in string format.
    """
    sign_string = "-" if seconds < 0 else ""
    seconds = abs(int(seconds))
    days, seconds = divmod(seconds, 86400)
    hours, seconds = divmod(seconds, 3600)
    minutes, seconds = divmod(seconds, 60)
    if days > 0:
        return "%s%dd%dh%dm" % (sign_string, days, hours, minutes)
    elif hours > 0:
        return "%s%dh%dm%ds" % (sign_string, hours, minutes, seconds)
    elif minutes > 0:
        return "%s%dm%ds" % (sign_string, minutes, seconds)
    else:
        return "%s%ds" % (sign_string, seconds)


def flatten_generator(items: Union[list, np.ndarray]) -> Generator:
    """
    Yield items from any nested iterable. This solution is modified from a recipe in
    Beazley, D. and B. Jones. Recipe 4.14, Python Cookbook 3rd Ed., O'Reilly Media Inc.
    Sebastopol, CA: 2013.

    Parameters
    ----------
    items
        A list, tuple, numpy.ndarray, etc. that should be flattened.

    Returns
    -------
    obj[generator]
        Can be translated to a list by applying list(...) on it.
    """
    for x in items:
        if type(x) in [list, tuple, np.ndarray] and not isinstance(x, (str, bytes)):
            for sub_x in flatten_generator(x):
                yield sub_x
        else:
            yield x


def flatten(arg: Union[list, np.ndarray, float, int, None]) -> Union[list, None]:
    """
    Flattens and returns the given input argument.

    Parameters
    ----------
    arg
        The list/array that should be flattened.

    Returns
    -------
    arg_flat
        The flattened list/numpy.ndarray is the input is not None. Otherwise, None is
        returned.
    """
    arg_type = type(arg)
    if arg is None:
        arg_flat = arg
    elif arg_type in [float, int]:
        arg_flat = [arg]
    elif arg_type in [list, np.ndarray]:
        arg_flat = list(flatten_generator(arg))  # type: ignore
    else:
        raise TypeError(
            f"The argument must be either None or of type list numpy.ndarray, "
            f"float or int. Found type '{arg_type}' however."
        )
    return arg_flat


def process_spatial_coordinates(
    x: Union[float, int, np.ndarray, None] = None,
    y: Union[float, int, np.ndarray, None] = None,
    z: Union[float, int, np.ndarray, None] = None,
    coords: Optional[np.ndarray] = None,
    order: tuple = ("x", "y", "z"),
) -> Tuple[np.ndarray, List[str]]:
    """
    Converts given spatial data from a flexible format into a standardized format.

    Parameters
    ----------
    x
        Positional x-coordinate. When given, the coords-argument must be None.
    y
        Positional y-coordinate. When given, the coords-argument must be None.
    z
        Positional z-coordinate. When given, the coords-argument must be None.
    coords
        Some or all of the coordinates x, y, z concatenated as an array. Each row
        corresponds to one coordinate. For example, row 1 might contain all the order-
        argument. When the coords-argument is given, all 3 args x, y and z must be None.
    order
        Only relevant when coords is given. Defines which row in coords corresponds to
        which coordinate. For example, order=('x', 'y', 'z') means that the 1st row are
        x-coordinates, the 2nd row are y-coords and the 3rd row are the z-coordinates.

    Returns
    -------
    coords
        An array with as many columns as coordinates are given, and as many rows as
        points are given. For example if 10 points with x and z coordinates are given,
        then coords would have a shape of (2, 10).
    adjusted_order
        Describes which coordinates are described by the rows of the returned coords.
        In the example given above, adjusted_order would be ['x', 'z'].
    """

    # the following check should cover the option that no spatial input is given
    if (x is None) and (y is None) and (z is None) and (coords is None):
        return np.array([]), []

    # convert all single-coordinate inputs to flat numpy arrays
    x = np.array(flatten(x)) if x is not None else None
    y = np.array(flatten(y)) if y is not None else None
    z = np.array(flatten(z)) if z is not None else None

    # derive the number of given coordinate vectors and points
    if coords is not None:
        if not type(coords) is np.ndarray:
            raise TypeError(
                f"The argument 'coords' must be of type numpy.ndarray. Found "
                f"{type(coords)} however."
            )
        else:
            # each row corresponds to one coordinate, so the number of given points is
            # the length of rows
            n_coords, n_points = coords.shape
    else:
        n_points_list = [len(v) for v in [x, y, z] if v is not None]
        n_points_set = set(n_points_list)
        if len(n_points_set) == 1:
            n_coords = len(n_points_list)
            n_points = n_points_list[0]
        else:
            raise RuntimeError(
                f"Found inconsistent lengths in given coordinate "
                f"vectors: {n_points_list}!"
            )

    # derive the coords array and the corresponding order-vector to be returned; note
    # that the repeated if-else clause here should improve readability
    if coords is not None:
        # it is assumed here that the first n_coords elements from the order-vector
        # correspond to the n_coords rows of the given coords-argument
        adjusted_order = list(order[:n_coords])
    else:
        # in this case the order-vector might have to be trimmed; for example if x and
        # z are given, the 'y' from the order vector has to be removed
        coords = np.zeros((n_coords, n_points))
        adjusted_order = []
        row_idx = 0
        for v in order:
            if eval(v) is not None:
                adjusted_order.append(v)
                coords[row_idx, :] = eval(v)
                row_idx += 1

    return coords, adjusted_order


def translate_prms_def(prms_def_given: Union[str, list, dict]) -> Tuple[dict, int]:
    """
    Translates the prms_def argument which is used by several sub-modules (e.g.
    ForwardModelBase, NoiseModelBase, PriorBase) into a default format. The prms_def-
    argument specifies the local/global names of the parameters used by a sub-module.

    Parameters
    ----------
    prms_def_given
        Either a single string, a dictionary with global names as keys and local names
        as values, or a list, the elements of which are either strings or 1-element
        dictionaries, where the latter would again contain one global name as key and
        one local name as value. Valid examples are: 'sigma', ['sigma'], ['sigma',
        'beta'], ['sigma', {'beta': 'b'}], {'sigma': 'sigma', 'beta': 'b'}.

    Returns
    -------
    prms_def
        Contains global names as keys and local names as values.
    prms_dim
        The number of items in prms_def.
    """
    prms_def_copy = copy(prms_def_given)
    if type(prms_def_copy) is dict:
        prms_def = list2dict(prms_def_copy)
    else:
        prms_def = list2dict(make_list(prms_def_copy))
    prms_dim = len(prms_def)
    return prms_def, prms_dim


def print_probeye_header(
    width: int = 100,
    header_file: str = "probeye.txt",
    version: str = "1.0.16",
    margin: int = 5,
    h_symbol: str = "=",
    v_symbol: str = "#",
    use_logger: bool = True,
):
    """
    Prints the probeye header which is printed, when an inference problem is set up.
    Mostly just nice to have. The only useful information it contains is the version
    number of the package.

    Parameters
    ----------
    width
        The width (i.e., number of characters) the header should have.
    header_file
        Relative path (with respect to this file) to the txt-file that contains the
        probeye letters.
    version
        States the probeye version; this should be identical to the version stated in
        setup.cfg; however, the version cannot be read dynamically, since the setup.cfg
        is not available after installing the package.
    margin
        Minimum number of blank spaces at the header margins.
    h_symbol
        The symbol used to 'draw' the horizontal frame line.
    v_symbol
        The symbol used to 'draw' the vertical frame line.
    use_logger
        When True, the header will be logged, otherwise just printed.
    """

    # define the full paths of the given files
    dir_path = os.path.dirname(__file__)
    header_file = os.path.join(dir_path, header_file)

    # read in the big probeye letters
    with open(header_file, "r") as f:
        content = f.readlines()
    # this is the width of the read in 'probeye' in terms of number of chars; note that
    # all lines (should) have the same length
    width_probeye = len(content[0]) - 1

    # this string should coincide with the one given in setup.cfg; however, it cannot be
    # read dynamically since the setup.cfg is not available after installing the package
    description = "A general framework for setting up parameter " "estimation problems."
    subtitle = f"Version {version} - {description}"
    width_subtitle = len(subtitle)

    # choose a width so that the margin on one side is at least 'margin'
    width_used = max(
        (width, width_probeye + 2 * (margin + 1), width_subtitle + 2 * margin + 1)
    )

    # assemble the header
    outer_frame_line = f"{v_symbol} {h_symbol * (width_used - 4)} {v_symbol}"
    inner_frame_line = f"{v_symbol}{' ' * (width_used - 2)}{v_symbol}"
    lines = [outer_frame_line, inner_frame_line]
    for line in content:
        clean_line = line.replace("\n", "")
        lines.append(f"{v_symbol}{clean_line:^{width_used - 2}s}{v_symbol}")
    lines.append(inner_frame_line)
    lines.append(outer_frame_line)
    lines.append(inner_frame_line)
    lines.append(f"{v_symbol}{subtitle:^{width_used - 2}s}{v_symbol}")
    lines.append(inner_frame_line)
    lines.append(outer_frame_line)

    # log or print the header
    if use_logger:
        print("")
        for line in lines:
            logger.info(line)
    else:
        print("\n" + "\n".join(lines))


def logging_setup(
    log_level_stdout: str = "INFO",
    log_level_file: str = "DEBUG",
    log_format: Optional[str] = None,
    log_file: Optional[str] = None,
    overwrite_log_file: bool = True,
    **kwargs,
):
    """
    Sets up the loguru logger for listening to the inference problem.

    Parameters
    ----------
    log_level_stdout
        Defines the level of the logging output to stdout. Common choices are 'DEBUG',
        'INFO', 'WARNING', and 'ERROR'.
    log_level_file
        Defines the level of the logging output to a log file. Common choices are again
        'DEBUG', 'INFO', 'WARNING', and 'ERROR'.
    log_format
        A format string defining the logging output. If this argument is
        set to None, a default format will be used.
    log_file
        Path to the log-file, if the logging should be printed to file. If
        None is given, no logging-file will be created.
    overwrite_log_file
        When True, a specified log-file will be overwritten. Otherwise,
        the generated logging will appended to a given log-file.
    kwargs
        Additional keyword arguments passed to logger.add (for file and stdout).
    """
    if not log_format:
        log_format = (
            "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
            "<level>{level: <8}</level> | "
            "<level>{message:100s}</level> | "
            "<cyan>{name}</cyan>:"
            "<cyan>{function}</cyan>:"
            "<cyan>{line}</cyan>"
        )
    logger.remove()  # just in case there still exists another logger
    logger.add(sys.stdout, format=log_format, level=log_level_stdout, **kwargs)
    if log_file:
        if os.path.isfile(log_file) and overwrite_log_file:
            os.remove(log_file)
        logger.add(log_file, format=log_format, level=log_level_file, **kwargs)


class StreamToLogger:
    """This class is required by stream_to_logger defined right below."""

    def __init__(self, level):
        self._level = level

    def write(self, buffer):
        for line in buffer.rstrip().splitlines():
            logger.opt(depth=1).log(self._level, line.rstrip())


def stream_to_logger(log_level: str) -> StreamToLogger:
    """
    Returns a stream-object that can be used to redirect a function's print output to
    the logger. Taken from the section 'Capturing standard stdout ...' of
    https://loguru.readthedocs.io/en/stable/resources/recipes.html.

    Parameters
    ----------
    log_level
        Defines the log level the streamed output will be associated with. Common
        choices are 'DEBUG', 'INFO', 'WARNING', and 'ERROR'.

    Returns
    -------
        This object should be used as follows:
        import contextlib
        with contextlib.redirect_stdout(stream_to_logger('INFO')):
            <function that prints something>
    """
    return StreamToLogger(log_level)


def print_dict_in_rows(
    d: dict, printer: Callable = print, sep: str = "=", val_fmt: Optional[str] = None
):
    """
    Prints a dictionary with key-value pairs in rows.

    Parameters
    ----------
    d
        The dictionary to print.
    printer
        Function used for printing. For example 'print' or 'logger.info'.
    sep
        The character printed between key and value.
    val_fmt
        A format string used for printing the dictionary's values.
    """
    n = max([len(key) for key in d.keys()])
    for key, val in d.items():
        if val_fmt:
            printer(f"{key:{n + 1}s} {sep} {val:{val_fmt}}")
        else:
            printer(f"{key:{n + 1}s} {sep} {val}")


def add_index_to_tex_prm_name(tex: str, index: int) -> str:
    """
    Adds a lower index to a parameter's tex-name. This function is intended for vector-
    valued parameters. For example: ('$a$', 1) -> '$a_1$'.

    Parameters
    ----------
    tex
        The tex-string to be modified.
    index
        The index to be added as a lower index to tex.

    Returns
    -------
    tex_mod
        The tex-string with included index.
    """

    # the math-model '$' should appear twice in the string
    check_1 = tex.count("$") == 2
    # the index is only added in tex-fashion
    # if no indexes are present already
    check_2 = not ("_" in tex)
    check_3 = not ("^" in tex)

    if check_1 and check_2 and check_3:
        tex_list = tex.split("$")
        # since it was checked that there are exactly 2 '$'-signs in tex, the tex_list
        # has 3 elements, with the middle one being the string enclosed by the two
        # '$'-signs
        tex_list[1] = tex_list[1] + f"_{index}"
        tex_mod = "$".join(tex_list)
    else:
        # if not all checks are passed, the index is added in a way, that does not
        # expect anything from the given tex-string
        tex_mod = tex + f" ({index})"

    return tex_mod


def check_for_uninformative_priors(problem: "InferenceProblem"):
    """
    Checks if all priors defined within a given InferenceProblem are not uninformative.

    Parameters
    ----------
    problem
        The given problem to check.
    """
    for prior_name, prior_template in problem.priors.items():
        if prior_template.prior_type == "uninformative":
            raise RuntimeError(
                f"The prior '{prior_name}' is uninformative, which cannot be used by "
                f"the requested solver. You could change it to a uniform-prior on a "
                f"specified interval to solver this problem."
            )
